from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from wise_pizza.utils import almost_equals


class TransformWithWeights(ABC):
    @abstractmethod
    def transform_mean(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_transform_mean(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def transform_weight(self, w: np.ndarray, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_transform_weight(self, w: np.ndarray, x: np.ndarray) -> np.ndarray:
        pass

    def transform_totals_weights(
        self, t: np.ndarray, w: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        mean = t / w
        t_mean = self.transform_mean(mean)
        t_w = self.transform_weight(w, mean)
        return t_mean * t_w, t_w

    def inverse_transform_totals_weights(
        self, t_total: np.ndarray, t_w: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        t_mean = t_total / t_w
        mean = self.inverse_transform_mean(t_mean)
        w = self.inverse_transform_weight(t_w, t_mean)
        return mean * w, w

    def test_transforms(self, total, weights, eps=1e-4):
        mean = total / weights
        t_mean = self.transform_mean(mean)
        assert almost_equals(mean, self.inverse_transform_mean(t_mean), eps)

        t_w = self.transform_weight(weights, mean)
        re_w = self.inverse_transform_weight(t_w, t_mean)
        assert almost_equals(weights, re_w, eps)

        t_t, tt_w = self.transform_totals_weights(total, weights)
        re_t, re_w = self.inverse_transform_totals_weights(t_t, tt_w)

        assert almost_equals(weights, re_w, eps)
        assert almost_equals(total, re_t, eps)


class IdentityTransform(TransformWithWeights):
    def transform_mean(self, x: np.ndarray) -> np.ndarray:
        return x

    def inverse_transform_mean(self, x: np.ndarray) -> np.ndarray:
        return x

    def transform_weight(self, w: np.ndarray, x: np.ndarray) -> np.ndarray:
        return w

    def inverse_transform_weight(self, w: np.ndarray, x: np.ndarray) -> np.ndarray:
        return w


class LogTransform(TransformWithWeights):
    def __init__(
        self, offset: float, weight_pow_sc: float = 0.1, cap_inverse: bool = True
    ):
        self.offset = offset
        self.weight_pow_sc = weight_pow_sc
        self.cap_inverse = cap_inverse
        if cap_inverse:
            self.max_inverse = 0.0
        else:
            self.max_inverse = None

    def transform_mean(self, x: np.ndarray) -> np.ndarray:
        if self.cap_inverse:
            self.max_inverse = np.maximum(self.max_inverse, 2 * x.max())
        return np.log(self.offset + x)

    def inverse_transform_mean(self, x: np.ndarray) -> np.ndarray:
        if self.cap_inverse:
            return np.maximum(
                0.0, np.exp(np.minimum(x, np.log(self.max_inverse))) - self.offset
            )
        else:
            return np.maximum(0.0, np.exp(x) - self.offset)

    def transform_weight(self, w: np.ndarray, mean: np.ndarray) -> np.ndarray:
        # pure math would give weight_pow_sc = 1, but then
        # there's too much information from actuals being leaked into the weights,
        # so the
        return w * np.power(self.offset + mean, self.weight_pow_sc)

    def inverse_transform_weight(
        self, t_w: np.ndarray, t_mean: np.ndarray
    ) -> np.ndarray:
        return t_w / np.power(
            self.offset + self.inverse_transform_mean(t_mean), self.weight_pow_sc
        )



import itertools
import os
from typing import Callable

import numpy as np
import pandas as pd
import pytest

from wise_pizza.data_sources.synthetic import synthetic_data, synthetic_ts_data
from wise_pizza.explain import (
    explain_changes_in_average,
    explain_changes_in_totals,
    explain_levels,
    explain_timeseries,
)
from wise_pizza.segment_data import SegmentData
from wise_pizza.solve.solver import solve_lasso, solve_lp
from wise_pizza.time import create_time_basis
from wise_pizza.plotting_time import plot_time

np.random.seed(42)

dims = [
    "TYPE",
    "REGION",
    "FIRST_PRODUCT",
    "CURRENCY",
]
totals = "VOLUME"
size = "TRANSACTION_COUNT"

# possible values for explain methods
# Too long, delete some values for quick starts, e.g. by deleting the parameters in nan_percent, size_one_percent
deltas_test_values = [
    ("totals", "split_fits", "force_dim", "extra_dim"),  # how
    ("lp", "lasso", "tree"),  # solver
    (True,),  # plot_is_static
    (explain_changes_in_average, explain_changes_in_totals),  # function
    (0.0, 90.0),  # nan_percent
    (0.0, 90.0),  # size_one_percent
    (True, False),  # cluster_values
]
# possible variants for explain methods
deltas_test_cases = list(itertools.product(*deltas_test_values))

# possible values for explain_levels
levels_test_values = [
    ("lp", "lasso", "tree"),  # solver
    (0.0, 90.0),  # nan_percent
    (0.0, 90.0),  # size_one_percent
]

# possible variants for explain_levels
levels_test_cases = list(itertools.product(*levels_test_values))


def values_to_nan(df: pd.DataFrame, nan_percent: float = 0.0):
    """
     Randomly change values to NaN in a pandas dataframe for testing.
    Parameters:
        - df: a pandas dataframe
        - percent: the percentage of values to replace with NaN, as a float between 0 and 100
    Returns:
        A new pandas dataframe with NaN values randomly changed.
    """
    df_copy = df.copy()  # create a copy of the original dataframe

    # calculate the number of values to replace with NaN
    num_values = int((nan_percent / 100) * len(df))
    np.random.seed(42)
    # get a random selection of indices to replace with NaN
    random_indices = df_copy.sample(num_values).index

    # replace values with NaN at the random indices
    df_copy.loc[random_indices, df_copy.columns] = np.nan

    return df_copy


def size_to_one(
    df: pd.DataFrame,
    percent: float = 0.0,
    size_name: str = "TRANSACTION_COUNT",
    totals_name: str = "VOLUME",
):
    """
     Randomly change sizes to one in a pandas dataframe for testing.
    Parameters:
        - df: a pandas dataframe
        - percent: the percentage of values to replace with one, as a float between 0 and 100
    Returns:
        A new pandas dataframe with changed sizes.
    """
    df_copy = df.copy()  # create a copy of the original dataframe

    # calculate the number of values to replace with NaN
    num_values = int((percent / 100) * len(df))
    np.random.seed(42)
    # get a random selection of indices to replace with NaN
    random_indices = df_copy.sample(num_values).index

    # replace values with NaN at the random indices
    df_copy.loc[random_indices, size_name] = 1
    df_copy.loc[random_indices, totals_name] = 1

    return df_copy


def monthly_driver_data():
    df = pd.read_csv(
        os.path.join(os.path.dirname(__file__), "../data", "synth_data.csv")
    )
    return SegmentData(
        data=df, dimensions=dims, segment_total=totals, segment_size=size
    )


def test_categorical():
    all_data = monthly_driver_data()
    df = all_data.data
    last_month = df.COHORT_MONTH.max()

    data = df[df.COHORT_MONTH == last_month]

    sf = explain_levels(
        data,
        dims=all_data.dimensions,
        total_name=all_data.segment_total,
        size_name=all_data.segment_size,
        max_depth=2,
    )
    print("***")
    for s in sf.segments:
        print(s)
    print(sf.summary())
    print(sf.relevant_cluster_names)
    print("yay!")


@pytest.mark.parametrize("nan_percent, clustering", [[0.0, False], [1.0, False]])
def test_synthetic_template(nan_percent: float, clustering: bool):
    all_data = synthetic_data(init_len=10000, dim_values=5)
    data = all_data.data

    data.loc[(data["dim0"] == 0) & (data["dim1"] == 1), "totals"] += 100
    data.loc[(data["dim1"] == 0) & (data["dim2"] == 1), "totals"] += 300

    if nan_percent > 0:
        data = values_to_nan(data, nan_percent)
    sf = explain_levels(
        data,
        dims=all_data.dimensions,
        total_name=all_data.segment_total,
        size_name=all_data.segment_size,
        max_depth=2,
        max_segments=5,
        verbose=1,
        solver="lp",
        cluster_values=clustering,
    )
    print("***")
    for s in sf.segments:
        print(s)

    assert abs(sf.segments[0]["coef"] - 300) < 2
    assert abs(sf.segments[1]["coef"] - 100) < 2

    # sf.plot()
    print("yay!")


@pytest.mark.parametrize("nan_percent", [0.0, 1.0])
def test_synthetic_template_tree(nan_percent: float):
    all_data = synthetic_data(init_len=1000)
    data = all_data.data

    data.loc[(data["dim0"] == 0) & (data["dim1"] == 1), "totals"] += 200
    data.loc[(data["dim1"] == 0) & (data["dim2"] == 1), "totals"] += 300

    if nan_percent > 0:
        data = values_to_nan(data, nan_percent)
    sf = explain_levels(
        data,
        dims=all_data.dimensions,
        total_name=all_data.segment_total,
        size_name=all_data.segment_size,
        max_depth=2,
        min_segments=5,
        verbose=1,
        solver="tree",
    )
    print("***")
    for s in sf.segments:
        print(s)

    # TODO: insert approppriate asserts
    # assert abs(sf.segments[0]["coef"] - 300) < 2
    # assert abs(sf.segments[1]["coef"] - 100) < 2

    # sf.plot()
    print("yay!")


@pytest.mark.parametrize("nan_percent", [0.0, 1.0])
def test_synthetic_ts_template(nan_percent: float):
    all_data = synthetic_ts_data(init_len=10000)

    # Add some big trends to the data
    # TODO: insert trend break patterns too
    months = np.array(sorted(all_data.data[all_data.time_col].unique()))
    basis = create_time_basis(months, baseline_dims=1)
    joined = pd.merge(all_data.data, basis, left_on="TIME", right_index=True)
    df = joined.drop(columns=basis.columns)

    loc1 = (df["dim0"] == 0) & (df["dim1"] == 1)
    loc2 = (df["dim1"] == 0) & (df["dim2"] == 1)

    df.loc[loc1, "totals"] += 100 * joined.loc[loc1, "Slope"]
    df.loc[loc2, "totals"] += 300 * joined.loc[loc2, "Slope"]

    if nan_percent > 0:
        df = values_to_nan(df, nan_percent)
    sf = explain_timeseries(
        df,
        dims=all_data.dimensions,
        total_name=all_data.segment_total,
        time_name=all_data.time_col,
        size_name=all_data.segment_size,
        max_depth=2,
        max_segments=5,
        verbose=True,
    )
    print("***")
    for s in sf.segments:
        print(s)

    plot_time(sf)

    assert abs(sf.segments[0]["coef"] - 300) < 2
    assert abs(sf.segments[1]["coef"] - 100) < 2

    # sf.plot()
    print("yay!")


@pytest.mark.parametrize(
    "how, solver, plot_is_static, function, nan_percent, size_one_percent, cluster_values",
    deltas_test_cases,
)
def test_deltas(
    how: str,
    solver: str,
    plot_is_static: bool,
    function: Callable,
    nan_percent: float,
    size_one_percent: float,
    cluster_values: bool,
):
    all_data = monthly_driver_data()
    df = all_data.data

    months = sorted(df.COHORT_MONTH.unique())

    data = df[df.COHORT_MONTH == months[-1]]
    pre_data = df[df.COHORT_MONTH == months[-2]]
    np.random.seed(42)
    pre_data = values_to_nan(pre_data, nan_percent)
    data = values_to_nan(data, nan_percent)
    pre_data = size_to_one(pre_data, size_one_percent)
    data = size_to_one(data, size_one_percent)

    sf = function(
        pre_data,
        data,
        all_data.dimensions,
        all_data.segment_total,
        all_data.segment_size,
        how=how,
        max_depth=1,
        max_segments=10,
        solver=solver,
        cluster_values=cluster_values
    )
    # sf.plot(plot_is_static=plot_is_static)
    print("yay!")


@pytest.mark.parametrize("solver, nan_percent, size_one_percent", levels_test_cases)
def test_explain_levels(solver: str, nan_percent: float, size_one_percent: float):
    all_data = monthly_driver_data()
    df = all_data.data

    months = sorted(df.COHORT_MONTH.unique())

    data = df[df.COHORT_MONTH == months[-1]]
    data = values_to_nan(data, nan_percent)
    data = size_to_one(data, size_one_percent)

    sf = explain_levels(
        df=data,
        dims=all_data.dimensions,
        total_name=all_data.segment_total,
        size_name=all_data.segment_size,
        max_depth=1,
        max_segments=10,
        solver=solver,
    )
    print(sf.summary())
    print("yay!")


def test_solve_lasso():
    X = pd.DataFrame(columns=["vol", "rev", "count"])
    X["vol"] = [-1, -2, -3]
    X["rev"] = [-4, -5, -6]
    X["count"] = [7, 8, 9]
    y = pd.DataFrame(columns=["target"])
    y["target"] = [10, 11, 12]
    lasso = solve_lasso(
        X,
        y,
        alpha=0.1,
        constrain_signs=True,
        drop_last_row=True,
    )
    assert len(lasso.coef_) == 3
    print("COEFS: ", lasso.coef_)


def test_solve_lp():
    X = pd.DataFrame(columns=["vol", "rev", "count"])
    X["vol"] = [-1, -2, -3]
    X["rev"] = [-4, -5, -6]
    X["count"] = [7, 8, 9]
    y = pd.DataFrame(columns=["target"])
    y["target"] = [10, 11, 12]
    X = np.copy(X)
    y = np.copy(y)
    lp = solve_lp(
        X,
        y,
        alpha=0.1,
        constrain_signs=True,
        drop_last_row=True,
    )
    assert len(lp.coef_) == 3
    print("COEFS: ", lp.coef_)



"""An example of Streamlit leveraging Wise pizza."""
# Streamlit part was initially developed by https://github.com/agusfigueroa-htg

import streamlit as st
import os, sys
import pandas as pd
from wise_pizza.explain import (
    explain_levels,
    explain_changes_in_totals,
    explain_changes_in_average,
)

def run_streamlit_app():

    # False if you want nice interactive plots
    # True if you want static plots (Doesn't work on all platforms yet)
    plot_is_static = False


    # SETTING PAGE CONFIG TO WIDE MODE AND ADDING A TITLE AND FAVICON
    st.set_page_config(layout="wide", page_title="Wise Pizza", page_icon=":pizza:")

    st.title("Wise Pizza powered by Streamlit")
    st.text(
        "Only categorical columns are accepted, bucket the numeric ones if you wanna use those"
    )

    # upload the file from the computer
    def load_data_upload():
        uploaded_file = st.file_uploader("Choose a file")
        if not uploaded_file:
            st.warning("Please input a dataset.")
            data = pd.DataFrame()
            st.stop()
        else:
            st.success("Dataset inputted.")
            data = pd.read_csv(uploaded_file)
        return data


    on = st.toggle("Use sample data from Github")
    url_data = r"https://raw.githubusercontent.com/transferwise/wise-pizza/main/data/synth_data.csv"

    # select the datasource, either local or from github
    if on:
        st.write(f"Downloading data from {url_data}!")
        df = pd.read_csv(url_data)
    else:
        df = load_data_upload()

    # show dataset preview
    st.text("Table preview")
    st.table(df.head(10))

    # ask the user for relevant dimensions
    dims = st.multiselect(
        "Select the dimensions you want to include in the analysis",
        df.select_dtypes(exclude=["number"]).columns.tolist(),
    )

    # ask the user via streamlit to select if they want to run a comparison between subgroups or not
    flag_comparison = st.toggle(
        "I want to run a comparison between two subgroups in my data"
    )

    # return columns that are candidate for comparison
    def flag_columns(df):
        binary_columns = df.columns[df.nunique() == 2].tolist()
        if not binary_columns:
            st.warning(
                "No column in the dataset is binary, no comparison can be carried out"
            )
            st.stop()
        return binary_columns


    if flag_comparison:
        # calculate binary columns
        binary_columns = flag_columns(df)

        # ask the user via streamlit for the flag column
        flag_column = st.selectbox(
            "What is the flag column of your dataset that defines the two subgroups?",
            binary_columns,
            index=None,
            placeholder="Select flag column column",
        )
        st.write("You selected:", flag_column)

        # wait until flag column is added
        if not flag_column:
            st.stop()

        # calculate unique flags in the dataset
        flags = sorted(df[flag_column].unique())

        # allow users to define what's "old" and "new" in the comparison
        flags_option = st.selectbox(
            "Which value in your flag column belongs to group A?", (flags)
        )


    # ask the user via streamlit for the target column
    totals = st.selectbox(
        "Name of column that contains totals per segment (e.g. GMV/revenue)",
        # display only numerical columns
        df.select_dtypes(include=["number"]).columns.tolist(),
        index=None,
        placeholder="Select target column",
    )
    st.write("You selected:", totals)

    # ask the user via streamlit for the size column
    size = st.selectbox(
        "Name of column containing segment size (e.g. number of users/number of transactions)",
        # display only numerical columns
        df.select_dtypes(include=["number"]).columns.tolist(),
        index=None,
        placeholder="Select volume column",
    )
    st.write("You selected:", size)

    if not totals or not size or not dims:
        st.warning("Please input all fields above.")
        st.stop()

    if not flag_comparison:
        st.subheader("Finding the juiciest slices", divider="rainbow")
        st.text("Find segments whose average is most different from the global one")
        ## finding juiciest slices
        solver = st.selectbox("Select a solver:", ("lasso", "lp"))
        min_segments = st.number_input(
            "Min segments: Minimum number of segments to find",
            min_value=1,
            max_value=200,
            value=20,
            step=1,
            key="min_segments",
        )
        max_segments = st.number_input(
            "Max segments: Maximum number of segments to find, defaults to min_segments",
            min_value=1,
            max_value=200,
            value=20,
            step=1,
            key="max_segments",
        )
        min_depth = st.number_input(
            "Min depth: Minimum number of dimension to constrain in segment definition",
            min_value=1,
            max_value=10,
            value=1,
            step=1,
            key="min_depth",
        )
        max_depth = st.number_input(
            "Max depth: Maximum number of dimension to constrain in segment definition",
            min_value=1,
            max_value=10,
            value=2,
            step=1,
            key="max_depth",
        )
        cluster_values = st.selectbox(
            "Cluster values: In addition to single-value slices, consider slices that consist of a group of segments from the same dimension with similar naive averages",
            (True, False),
        )

        width = st.number_input(
            "Width of the plot",
            min_value=100,
            max_value=10000,
            value=500,
            step=1,
            key="width",
        )
        height = st.number_input(
            "Height of the plot",
            min_value=100,
            max_value=10000,
            value=500,
            step=1,
            key="height",
        )
        apply_button = st.button("Apply")
        if apply_button:
            sf = explain_levels(
                df=df,
                dims=dims,
                total_name=totals,
                size_name=size,
                min_depth=min_depth,
                max_depth=max_depth,
                min_segments=min_segments,
                max_segments=max_segments,
                solver=solver,
                cluster_values=cluster_values,
            )
            plot_sf = sf.plot(width=width, height=height, return_fig=True)
            if not cluster_values:
                plot_sf_0 = plot_sf
                st.plotly_chart(plot_sf_0, use_container_width=True)
            else:
                plot_sf_0 = plot_sf[0]
                st.plotly_chart(plot_sf_0, use_container_width=True)
                plot_sf_1 = plot_sf[1]
                st.plotly_chart(plot_sf_1, use_container_width=True)

    if flag_comparison:
        st.subheader("Analysing differences", divider="rainbow")
        st.text(
            "This section does compare the two groups defined by the flag. Old total is the group A you selected in the dropdown"
        )
        st.text(
            "Explain changes in totals: Let's look for segments that experience the largest change in the totals from previous period."
        )
        st.text(
            "Explain changes in average: Sometimes, rather than explaining the change in totals from one period to the next, one wishes to explain a change in averages"
        )

        data = df[df[flag_column] != flags_option]  # take the group to compare to
        pre_data = df[df[flag_column] == flags_option]  # take the group to be compared

        comparison_dimensions = list(filter(lambda x: x != flag_column, dims))

        explain = st.selectbox(
            "Select a function:",
            ("explain changes in totals", "explain changes in average"),
        )
        solver = st.selectbox("Select a solver:", ("lasso", "lp"))
        min_segments = st.number_input(
            "Min segments: Minimum number of segments to find",
            min_value=1,
            max_value=200,
            value=20,
            step=1,
            key="min_segments",
        )
        max_segments = st.number_input(
            "Max segments: Maximum number of segments to find, defaults to min_segments",
            min_value=1,
            max_value=200,
            value=20,
            step=1,
            key="max_segments",
        )
        min_depth = st.number_input(
            "Min depth: Minimum number of dimension to constrain in segment definition",
            min_value=1,
            max_value=10,
            value=1,
            step=1,
            key="min_depth",
        )
        max_depth = st.number_input(
            "Max depth: Maximum number of dimension to constrain in segment definition",
            min_value=1,
            max_value=10,
            value=2,
            step=1,
            key="max_depth",
        )
        cluster_values = st.selectbox(
            "Cluster values: In addition to single-value slices, consider slices that consist of a group of segments from the same dimension with similar naive averages",
            (True, False),
        )
        how = st.selectbox(
            "Select a method:", ("totals", "split_fits", "extra_dim", "force_dim")
        )
        st.text(
            '"totals": to only decompose segment totals (ignoring size vs average contribution)'
        )
        st.text(
            '"split_fits": to separately decompose contribution of size changes and average changes'
        )
        st.text(
            '"extra_dim": to treat size vs average change contribution as an additional dimension'
        )
        st.text(
            '"force_dim": like extra_dim, but each segment must contain a Change_from constraint'
        )
        width = st.number_input(
            "Width of the plot",
            min_value=100,
            max_value=10000,
            value=500,
            step=1,
            key="width",
        )
        height = st.number_input(
            "Height of the plot",
            min_value=100,
            max_value=10000,
            value=500,
            step=1,
            key="height",
        )
        changes_apply_button = st.button("Get results")
        ## running explain calculations
        if changes_apply_button:
            if explain == "explain changes in totals":
                sf1 = explain_changes_in_totals(
                    df1=pre_data,
                    df2=data,
                    dims=comparison_dimensions,
                    total_name=totals,
                    size_name=size,
                    min_depth=min_depth,
                    max_depth=max_depth,
                    min_segments=min_segments,
                    max_segments=max_segments,
                    solver=solver,
                    cluster_values=cluster_values,
                    how=how,
                )
            else:
                sf1 = explain_changes_in_average(
                    df1=pre_data,
                    df2=data,
                    dims=comparison_dimensions,
                    total_name=totals,
                    size_name=size,
                    min_depth=min_depth,
                    max_depth=max_depth,
                    min_segments=min_segments,
                    max_segments=max_segments,
                    solver=solver,
                    cluster_values=cluster_values,
                    how=how,
                )
            plot_sf1 = sf1.plot(width=width, height=height, return_fig=True)
            if not cluster_values:
                plot_sf_0 = plot_sf1
                st.plotly_chart(plot_sf_0, use_container_width=True)
            else:
                plot_sf_0 = plot_sf1[0]
                st.plotly_chart(plot_sf_0, use_container_width=True)
                plot_sf_1 = plot_sf1[1]
                st.plotly_chart(plot_sf_1, use_container_width=True)

if __name__ == "__main__":
    run_streamlit_app()



import copy
from typing import Optional, List, Dict, Tuple

import numpy as np
import pandas as pd
from scipy.sparse import csc_matrix

from .weighted_quantiles import weighted_quantiles
from .fitter import AverageFitter, Fitter
from wise_pizza.cluster import nice_cluster_names


def tree_solver(
    dim_df: pd.DataFrame,
    dims: List[str],
    time_basis: Optional[pd.DataFrame] = None,
    max_depth: Optional[int] = None,
    num_leaves: Optional[int] = None,
):
    if time_basis is None:
        fitter = AverageFitter()
    else:
        raise NotImplementedError("Time fitter not yet implemented")
        # fitter = TimeFitter(dims, list(time_basis.columns))

    df = dim_df.copy().reset_index(drop=True)
    df["__avg"] = df["totals"] / df["weights"]
    df["__avg"] = df["__avg"].fillna(df["__avg"].mean())

    root = ModelNode(df=df, fitter=fitter, dims=dims, max_depth=max_depth)

    build_tree(root=root, num_leaves=num_leaves, max_depth=max_depth)

    leaves = get_leaves(root)

    col_defs, cluster_names = nice_cluster_names([leaf.dim_split for leaf in leaves])

    for l, leaf in enumerate(leaves):
        leaf.df["Segment_id"] = l

    re_df = pd.concat([leaf.df for leaf in leaves]).sort_values(dims)
    X = pd.get_dummies(re_df["Segment_id"]).values

    return csc_matrix(X), col_defs, cluster_names


def error(x: np.ndarray, y: np.ndarray) -> float:
    return np.sum((x - y) ** 2)


def target_encode(df: pd.DataFrame, dim: str) -> dict:
    df = df[[dim, "totals", "weights"]]
    agg = df.groupby(dim, as_index=False).sum()
    agg["__avg"] = agg["totals"] / agg["weights"]
    agg["__avg"] = agg["__avg"].fillna(agg["__avg"].mean())
    enc_map = {k: v for k, v in zip(agg[dim], agg["__avg"])}

    if np.isnan(np.array(list(enc_map.values()))).any():
        raise ValueError("NaNs in encoded values")
    return enc_map


class ModelNode:
    def __init__(
        self,
        df: pd.DataFrame,
        fitter: Fitter,
        dims: List[str],
        max_depth: Optional[int] = None,
        dim_split: Optional[Dict[str, List]] = None,
    ):
        self.df = df.copy()
        self.fitter = fitter
        self.dims = dims
        self.max_depth = max_depth
        self._best_submodels = None
        self._error_improvement = float("-inf")
        self.children = None
        self.dim_split = dim_split or {}
        self.model = None

    @property
    def depth(self):
        return len(self.dim_split)

    @property
    def error(self):
        if self.model is None:
            self.model = copy.deepcopy(self.fitter)
            self.model.fit(
                X=self.df[self.dims],
                y=self.df["__avg"],
                sample_weight=self.df["weights"],
            )
        return self.model.error(
            self.df[self.dims], self.df["__avg"], self.df["weights"]
        )

    @property
    def error_improvement(self):
        if self.max_depth is None:
            self.max_depth = float("inf")
        if self._best_submodels is None:
            best_error = float("inf")

            if self.depth > self.max_depth:
                raise ValueError("Max depth exceeded")
            elif self.depth == self.max_depth:
                iter_dims = list(self.dim_split.keys())
            else:
                iter_dims = self.dims

            for dim in iter_dims:
                if len(self.df[dim].unique()) == 1:
                    continue
                enc_map = target_encode(self.df, dim)
                self.df[dim + "_encoded"] = self.df[dim].apply(lambda x: enc_map[x])
                if np.any(np.isnan(self.df[dim + "_encoded"])):  # pragma: no cover
                    raise ValueError("NaNs in encoded values")
                # Get split candidates for brute force search
                deciles = np.array([q / 10.0 for q in range(1, 10)])

                splits = weighted_quantiles(
                    self.df[dim + "_encoded"], deciles, self.df["weights"]
                )

                for split in np.unique(splits):
                    left = self.df[self.df[dim + "_encoded"] < split]
                    right = self.df[self.df[dim + "_encoded"] >= split]
                    if len(left) == 0 or len(right) == 0:
                        continue
                    dim_values1 = [k for k, v in enc_map.items() if v < split]
                    dim_values2 = [k for k, v in enc_map.items() if v >= split]
                    left_candidate = ModelNode(
                        df=left,
                        fitter=self.fitter,
                        dims=self.dims,
                        dim_split={**self.dim_split, **{dim: dim_values1}},
                        max_depth=self.max_depth,
                    )
                    right_candidate = ModelNode(
                        df=right,
                        fitter=self.fitter,
                        dims=self.dims,
                        dim_split={**self.dim_split, **{dim: dim_values2}},
                        max_depth=self.max_depth,
                    )

                    err = left_candidate.error + right_candidate.error
                    if err < best_error:
                        best_error = err
                        self._error_improvement = self.error - best_error
                        self._best_submodels = (left_candidate, right_candidate)

        return self._error_improvement


def get_best_subtree_result(
    node: ModelNode, max_depth: Optional[int] = 1000
) -> ModelNode:
    if node.children is None or node.depth >= max_depth:
        return node
    else:
        node1 = get_best_subtree_result(node.children[0])
        node2 = get_best_subtree_result(node.children[1])
        improvement1 = node1.error_improvement
        improvement2 = node2.error_improvement
        if improvement1 > improvement2:
            return node1
        else:
            return node2


def build_tree(root: ModelNode, num_leaves: int, max_depth: Optional[int] = 1000):
    for _ in range(num_leaves - 1):
        best_node = get_best_subtree_result(root, max_depth)
        if best_node.error_improvement > 0:
            best_node.children = best_node._best_submodels
        else:
            break


def get_leaves(node: ModelNode) -> List[ModelNode]:
    if node.children is None:
        return [node]
    else:
        return get_leaves(node.children[0]) + get_leaves(node.children[1])



import pandas as pd
from scipy.sparse import csc_matrix, hstack
from typing import Optional
import numpy as np
# import numba

class HeuristicSelector:
    def __init__(
        self,
        weights: np.ndarray,
        totals: np.ndarray,
        max_cols: int = 300,
        time_basis: Optional[pd.DataFrame] = None,
        verbose: bool = False,
    ):
        self.max_cols = max_cols
        self.col_defs = []
        self.weights = weights.reshape(-1, 1)
        self.totals = totals.reshape(-1, 1)
        self.time_basis = time_basis
        self.tvalues = np.array([])
        self.impact = np.array([])
        self.X = None
        self.verbose = verbose

    def __call__(self, nextX, col_defs):
        if self.verbose:
            print(f"Preprocessor received X: {nextX.shape[0]}x{nextX.shape[1]}")
        assert len(col_defs) == nextX.shape[1]

        tvalues, impact = get_metrics_new(nextX, self.totals, self.weights)

        self.tvalues = np.concatenate([self.tvalues, tvalues])
        self.impact = np.concatenate([self.impact, impact])
        self.col_defs += col_defs
        self.X = nextX if self.X is None else hstack([self.X, nextX])

        assert len(self.col_defs) == self.X.shape[1]

        if self.X.shape[1] > self.max_cols:
            chunk_size = int(2 * self.max_cols / 3)

            inds = []

            unusual = np.argsort(np.abs(self.tvalues))
            inds += list(unusual[-chunk_size:])

            unusual2 = np.argsort(np.abs(self.impact))
            inds += list(unusual2[-chunk_size:])

            best = np.array(list(set(inds)))

            self.tvalues = np.array(self.tvalues[best])
            self.impact = np.array(self.impact[best])
            self.X = csc_matrix(self.X[:, best])
            self.col_defs = [self.col_defs[i] for i in best]
            assert len(self.col_defs) == self.X.shape[1]
            del nextX, col_defs
        if self.verbose:
            print(f"Done preprocessing that one!")
        return self.X, self.col_defs



def get_metrics_new(X, y, w):
    # y is totals, so we're approximating y by WX, unweighted
    X = X.toarray()
    WX = w * X

    # Do a weighted regression **on each col of X separately**
    XtWy = WX.T.dot(y).T
    XtWX = (WX * WX).sum(axis=0, keepdims=True)
    coeffs = XtWy / XtWX
    sigmasq=get_sigmasq(coeffs, WX, y)
    stds = np.sqrt(sigmasq / XtWX)

    # One way of choosing "good" candidates is individually good t-values
    tvalues = coeffs / stds
    # Another is the total absolute difference in totals that this regressor would predict alone
    impact = np.abs(WX).sum(axis=0, keepdims=True) * coeffs
    return tvalues.reshape(-1), impact.reshape(-1)

# This function for some reason is by far the slowest part of the code
# JIT didn't seem to help as all the calc is in the numpy binaries anyway
# @numba.jit(nopython=True)  # Enable the JIT compiler with nopython mode for best performance
def get_sigmasq(coeffs: np.ndarray, WX: np.ndarray, y: np.ndarray) -> np.ndarray:
    err = coeffs * WX - y
    sigmasq = (err * err).sum(axis=0)
    return sigmasq.reshape(1,-1)

def get_metrics(X, y, w):
    w2 = w * w
    X = X.toarray()
    W2X = w2 * X
    # Do a weighted regression **on each col of X separately**
    XtWy = W2X.T.dot(y).T
    XtWX = (W2X * X).sum(axis=0, keepdims=True)
    coeffs = XtWy / XtWX
    err = coeffs * X - y
    sigmasq = (err * w * err).sum(axis=0, keepdims=True)
    stds = np.sqrt(sigmasq / XtWX)

    # One way of choosing "good" candidates is individually good t-values
    tvalues = coeffs / stds
    # Another is the total absolute difference in totals that this regressor would predict alone
    impact = np.abs(W2X).sum(axis=0, keepdims=True) * coeffs
    return tvalues.reshape(-1), impact.reshape(-1)


import copy
import warnings
from typing import List, Optional

import numpy as np
import pandas as pd

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

from wise_pizza.plotting import (
    plot_segments,
    plot_split_segments,
    plot_waterfall,
)
from wise_pizza.plotting_time import plot_time, plot_ts_pair
from wise_pizza.slicer import SliceFinder, SlicerPair
from wise_pizza.slicer_facades import TransformedSliceFinder
from wise_pizza.utils import diff_dataset, prepare_df, almost_equals
from wise_pizza.time import create_time_basis, add_average_over_time, extend_dataframe
from wise_pizza.transform import IdentityTransform, LogTransform


def explain_changes_in_average(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: str,
    min_segments: Optional[int] = None,
    max_segments: int = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "lasso",
    how: str = "totals",
    force_add_up: bool = False,
    constrain_signs: bool = True,
    cluster_values: bool = False,
    verbose: int = 0,
):
    """
    Find segments most useful in explaining the difference between the averages of the two datasets
    @param df1: First dataset
    @param df2: Second dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param how: "totals" to only decompose segment totals (ignoring size vs average contribution)
            "split_fits" to separately decompose contribution of size changes and average changes
            "extra_dim" to treat size vs average change contribution as an additional dimension
            "force_dim" like extra_dim, but each segment must contain a Change_from constraint
    @param force_add_up: Force the contributions of chosen segments to add up
    to the difference between dataset totals
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param cluster_values: In addition to single-value slices, consider slices that consist of a
    group of segments from the same dimension with similar naive averages
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @return: A fitted object
    """
    df1 = df1.copy()
    df2 = df2.copy()

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df1 = prepare_df(df1, dims, size_name, total_name)
    df2 = prepare_df(df2, dims, size_name, total_name)

    # rescale sizes and totals, preserving averages (= total/size)
    df1["Norm_weight"] = df1[size_name] / df1[size_name].sum()
    df2["Norm_weight"] = df2[size_name] / df2[size_name].sum()

    df1["Norm_totals"] = df1[total_name] / df1[size_name].sum()
    df2["Norm_totals"] = df2[total_name] / df2[size_name].sum()

    # subtract the initial average from both totals
    avg1 = df1["Norm_totals"].sum()
    df1["Adj_totals"] = df1["Norm_totals"] - avg1 * df1["Norm_weight"]
    df2["Adj_totals"] = df2["Norm_totals"] - avg1 * df2["Norm_weight"]

    # call explain_changes
    sf = explain_changes_in_totals(
        df1,
        df2,
        dims,
        total_name="Adj_totals",
        size_name="Norm_weight",
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        how=how,
        force_add_up=force_add_up,
        constrain_signs=constrain_signs,
        cluster_values=cluster_values,
        verbose=verbose,
    )

    if hasattr(sf, "pre_total"):
        sf.pre_total = avg1
        sf.post_total += avg1
        sfs = None
    # Want to put the subtracted avg1 back in, something like
    # for s in sf.segments:
    #     s["naive_avg"] += average
    #     s["total"] += average * s["seg_size"]
    # print(average)
    # sf.reg.intercept_ = average

    # And might want to relabel some plots?
    sf.task = "changes in average"
    return sf


def explain_changes_in_totals(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: str,
    min_segments: Optional[int] = None,
    max_segments: int = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "lasso",
    how: str = "totals",
    force_add_up: bool = False,
    constrain_signs: bool = True,
    cluster_values: bool = False,
    verbose: int = 0,
):
    """
    Find segments most useful in explaining the difference between the totals of the two datasets
    @param df1: First dataset
    @param df2: Second dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param how: "totals" to only decompose segment totals (ignoring size vs average contribution)
            "split_fits" to separately decompose contribution of size changes and average changes
            "extra_dim" to treat size vs average change contribution as an additional dimension
            "force_dim" like extra_dim, but each segment must contain a Change_from constraint
    @param force_add_up: Force the contributions of chosen segments to add up
    to the difference between dataset totals
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param cluster_values: In addition to single-value slices, consider slices that consist of a
    group of segments from the same dimension with similar naive averages
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @return: A fitted object
    """

    assert how in ["totals", "extra_dim", "split_fits", "force_dim"]
    split_deltas = not (how == "totals")
    return_multiple = how == "split_fits"
    final_size = df2[size_name].sum()

    df1 = df1.copy()
    df2 = df2.copy()

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df1 = prepare_df(df1, dims, size_name, total_name)
    df2 = prepare_df(df2, dims, size_name, total_name)

    my_diff = diff_dataset(
        df1,
        df2,
        dims,
        total_name,
        size_name,
        split_deltas=split_deltas,
        return_multiple=return_multiple,
    )

    if how == "split_fits":
        df_size, df_avg = my_diff
        sf_size = explain_levels(
            df=df_size.data,
            dims=dims,
            total_name=df_size.segment_total,
            size_name=df_size.segment_size,
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            solver=solver,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            cluster_values=cluster_values,
            verbose=verbose,
        )

        sf_avg = explain_levels(
            df=df_avg.data,
            dims=dims,
            total_name=df_avg.segment_total,
            size_name=df_avg.segment_size,
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            solver=solver,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            cluster_values=cluster_values,
            verbose=verbose,
        )

        sf_size.final_size = final_size
        sf_avg.final_size = final_size
        sp = SlicerPair(sf_size, sf_avg)
        sp.plot = lambda plot_is_static=False, width=2000, height=500, cluster_key_width=180, cluster_value_width=318, return_fig=False: plot_split_segments(
            sp.s1,
            sp.s2,
            plot_is_static=plot_is_static,
            width=width,
            height=height,
            cluster_key_width=cluster_key_width,
            cluster_value_width=cluster_value_width,
            return_fig=return_fig,
        )
        return sp

    else:
        sf = SliceFinder()

        sf.fit(
            my_diff.data[my_diff.dimensions],
            my_diff.data[my_diff.segment_total],
            weights=my_diff.data[my_diff.segment_size],
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            max_segments=max_segments,
            solver=solver,
            force_dim="Change from" if how == "force_dim" else None,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            cluster_values=cluster_values,
            verbose=verbose,
        )

        sf.pre_total = df1[total_name].sum()
        sf.post_total = df2[total_name].sum()

        sf.plot = lambda plot_is_static=False, width=1000, height=1000, cluster_key_width=180, cluster_value_width=318, return_fig=False: plot_waterfall(
            sf,
            plot_is_static=plot_is_static,
            width=width,
            height=height,
            cluster_key_width=cluster_key_width,
            cluster_value_width=cluster_value_width,
            return_fig=return_fig,
        )
        sf.task = "changes in totals"
        return sf


def explain_levels(
    df: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: Optional[str] = None,
    min_segments: int = None,
    max_segments: int = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver="lasso",
    verbose=0,
    force_add_up: bool = False,
    constrain_signs: bool = True,
    cluster_values: bool = False,
):
    """
    Find segments whose average is most different from the global one
    @param df: Dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @param force_add_up: Force the contributions of chosen segments to add up to zero
    @param constrain_signs: Whether to constrain weights of segments to have the same sign as naive segment averages
    @param cluster_values: In addition to single-value slices, consider slices that consist of a
    group of segments from the same dimension with similar naive averages
    @return: A fitted object
    """
    df = copy.copy(df)

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df = prepare_df(df, dims, size_name, total_name)

    if size_name is None:
        size_name = "size"
        df[size_name] = 1.0

    # we want to look for deviations from average value
    average = df[total_name].sum() / df[size_name].sum()
    df["_target"] = df[total_name] - df[size_name] * average

    sf = SliceFinder()
    sf.fit(
        df[dims],
        df["_target"],
        weights=None if size_name is None else df[size_name],
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        verbose=verbose,
        force_add_up=force_add_up,
        constrain_signs=constrain_signs,
        cluster_values=cluster_values,
    )

    for s in sf.segments:
        s["naive_avg"] += average
        s["total"] += average * s["seg_size"]
    # print(average)
    sf.reg.intercept_ = average
    sf.plot = lambda plot_is_static=False, width=2000, height=500, return_fig=False, cluster_key_width=180, cluster_value_width=318: plot_segments(
        sf,
        plot_is_static=plot_is_static,
        width=width,
        height=height,
        return_fig=return_fig,
        cluster_key_width=cluster_key_width,
        cluster_value_width=cluster_value_width,
    )
    sf.task = "levels"
    return sf


def explain_timeseries(
    df: pd.DataFrame,
    dims: List[str],
    total_name: str,
    time_name: str,
    size_name: Optional[str] = None,
    min_segments: int = None,
    max_segments: int = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "omp",
    verbose: bool = False,
    constrain_signs: bool = False,
    cluster_values: bool = False,
    time_basis: Optional[pd.DataFrame] = None,
    fit_log_space: bool = False,
    fit_sizes: Optional[bool] = None,
    log_space_weight_sc: float = 0.5,
):
    df = copy.copy(df)

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    # Group by dims + [time_name]
    df = prepare_df(
        df, dims, total_name=total_name, size_name=size_name, time_name=time_name
    )
    df = df.sort_values(by=dims + [time_name])

    if size_name is None:
        size_name = "size"
        df[size_name] = 1.0
        if fit_sizes == True:
            raise ValueError("fit_sizes should be None or False if size_name is None")
        fit_sizes = False
    else:
        if fit_sizes is None:
            fit_sizes = True

    if fit_log_space:
        tf = LogTransform(
            offset=1,
            weight_pow_sc=log_space_weight_sc,
        )
    else:
        tf = IdentityTransform()

    size_name_orig = size_name + "_orig"
    total_name_orig = total_name + "_orig"

    df2 = df.rename(columns={size_name: size_name_orig, total_name: total_name_orig})

    if not fit_sizes:
        t, w = tf.transform_totals_weights(
            df2[total_name_orig].values, df2[size_name_orig].values
        )
        df2[total_name] = pd.Series(data=t, index=df2.index)
        df2[size_name] = pd.Series(data=w, index=df2.index)
        sf_totals = _explain_timeseries(
            df=df2,
            dims=dims,
            total_name=total_name,
            time_name=time_name,
            size_name=size_name,
            min_segments=min_segments,
            max_segments=max_segments,
            min_depth=min_depth,
            max_depth=max_depth,
            solver=solver,
            verbose=verbose,
            constrain_signs=constrain_signs,
            cluster_values=cluster_values,
            time_basis=time_basis,
        )
        return TransformedSliceFinder(sf_totals, transformer=tf)

    this_w = np.ones_like(df2[size_name_orig].values)
    these_totals = df2[size_name_orig].values

    tf.test_transforms(this_w, these_totals)

    t, w = tf.transform_totals_weights(these_totals, this_w)
    df2[size_name] = pd.Series(data=t, index=df2.index)
    df2["resc_wgt"] = pd.Series(data=w, index=df2.index)

    sf_wgt = _explain_timeseries(
        df=df2,
        dims=dims,
        total_name=size_name,
        size_name="resc_wgt",
        time_name=time_name,
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        verbose=verbose,
        constrain_signs=constrain_signs,
        cluster_values=cluster_values,
        time_basis=time_basis,
    )

    sf1 = TransformedSliceFinder(sf_wgt, transformer=tf)

    # Replace actual weights with fitted ones, for consistent extrapolation
    eps = 1e-3
    fitted_sizes = np.maximum(sf1.predicted_totals, eps)
    fitted_sizes[np.isnan(fitted_sizes)] = eps
    actual_avgs = df2[total_name_orig].values / df2[size_name_orig].values
    adj_totals = actual_avgs * fitted_sizes

    if fit_log_space:
        tf2 = LogTransform(offset=1, weight_pow_sc=log_space_weight_sc)
    else:
        tf2 = IdentityTransform()

    tf2.test_transforms(adj_totals, fitted_sizes)

    t, w = tf2.transform_totals_weights(adj_totals, fitted_sizes)
    df2[total_name] = pd.Series(data=t, index=df2.index)
    df2[size_name] = pd.Series(data=w, index=df2.index)

    sf_totals = _explain_timeseries(
        df=df2,
        dims=dims,
        total_name=total_name,
        time_name=time_name,
        size_name=size_name,
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        verbose=verbose,
        constrain_signs=constrain_signs,
        cluster_values=cluster_values,
        time_basis=time_basis,
    )

    assert almost_equals(t, sf_totals.actual_totals)
    assert almost_equals(w, sf_totals.weights)

    sf2 = TransformedSliceFinder(sf_totals, tf2)

    assert almost_equals(adj_totals, sf2.actual_totals)
    assert almost_equals(fitted_sizes, sf2.weights)

    out = SlicerPair(sf1, sf2)
    out.relevant_cluster_names1 = sf1.relevant_cluster_names
    out.relevant_cluster_names2 = sf2.relevant_cluster_names

    out.plot = lambda plot_is_static=False, width=600, height=1200, return_fig=False, average_name=None, use_fitted_weights=False: plot_ts_pair(
        out.s1,
        out.s2,
        plot_is_static=plot_is_static,
        width=width,
        height=height,
        return_fig=return_fig,
        average_name=average_name,
        use_fitted_weights=use_fitted_weights,
    )
    out.task = "time with weights"
    return out


def _explain_timeseries(
    df: pd.DataFrame,
    dims: List[str],
    total_name: str,
    time_name: str,
    size_name: Optional[str] = None,
    min_segments: int = None,
    max_segments: int = 5,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "omp",
    verbose: bool = False,
    force_add_up: bool = False,
    constrain_signs: bool = False,
    cluster_values: bool = False,
    time_basis: Optional[pd.DataFrame] = None,
):
    """
    Find segments whose average is most different from the global one
    @param df: Dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param time_name: Name of column containing the time dimension
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @param force_add_up: Force the contributions of chosen segments to add up to zero
    @param constrain_signs: Whether to constrain weights of segments to have the same sign as naive segment averages
    @param cluster_values In addition to single-value slices, consider slices that consist of a
    group of segments from the same dimension with similar naive averages
    @return: A fitted object
    """

    # strip out constants and possibly linear trends for each dimension combination
    baseline_dims = 1
    if time_basis is None:
        time_basis = create_time_basis(
            df[time_name].unique(), baseline_dims=baseline_dims, include_breaks=True
        )
        dtrend_cols = [t for t in time_basis.columns if "dtrend" in t]
        chosen_cols = []
        num_breaks = 2
        for i in range(1, num_breaks + 1):
            chosen_cols.append(
                dtrend_cols[int(i * len(dtrend_cols) / (num_breaks + 1))]
            )
        pre_basis = time_basis[list(time_basis.columns[:2]) + chosen_cols].copy()
        # TODO: fix this bug
        for c in chosen_cols:
            pre_basis[c + "_a"] = pre_basis["Slope"] - pre_basis[c]

        # print("yay!")

    df, avg_df = add_average_over_time(
        df,
        dims=dims,
        total_name=total_name,
        size_name=size_name,
        time_name=time_name,
        cartesian=False,
    )
    # The join in the above function could have messed up the ordering
    df = df.sort_values(by=dims + [time_name])

    # This block is pointless as we just normalized each sub-segment to zero average across time
    average = df[total_name].sum() / df[size_name].sum()
    df["_target"] = df[total_name] - df["total_adjustment"]

    sf = SliceFinder()
    sf.global_average = average
    sf.total_name = total_name
    sf.size_name = size_name
    sf.time_name = time_name
    sf.y_adj = df["total_adjustment"].values
    sf.avg_df = avg_df
    sf.time_values = df[time_name].unique()
    sf.fit(
        df[dims],
        df["_target"],
        time_col=df[time_name],
        time_basis=pre_basis,
        weights=df[size_name],
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        verbose=verbose,
        force_add_up=force_add_up,
        constrain_signs=constrain_signs,
        cluster_values=cluster_values,
    )

    # TODO: insert back the normalized bits?
    for s in sf.segments:
        segment_def = s["segment"]
        assert "time" in segment_def, "Each segment should have a time profile!"
        this_vec = (
            sf.X[:, s["index"]]
            .toarray()
            .reshape(
                -1,
            )
        )
        time_mult = (
            sf.time_basis[segment_def["time"]]
            .toarray()
            .reshape(
                -1,
            )
        )
        dummy = (this_vec / time_mult).astype(int).astype(np.float64)
        s["dummy"] = dummy
        s["seg_total_vec"] = this_vec * s["coef"] * sf.weights
        if len(segment_def) > 1:
            elems = np.unique(dummy)
            assert len(elems) == 2
            assert 1.0 in elems
            assert 0.0 in elems

        s["naive_avg"] += average
        s["total"] += average * s["seg_size"]
    # print(average)
    # sf.reg.intercept_ += average
    sf.plot = lambda plot_is_static=False, width=1200, height=2000, return_fig=False, average_name=None: plot_time(
        sf,
        plot_is_static=plot_is_static,
        width=width,
        height=height,
        return_fig=return_fig,
        average_name=average_name,
    )
    sf.task = "time"
    return sf



import copy
import itertools
from typing import Optional, List, Dict, Sequence
from collections import defaultdict
import math

import numpy as np
import scipy
from tqdm import tqdm
import pandas as pd
from scipy.sparse import csc_matrix, hstack, diags


def join_to_sparse(dim_df: pd.DataFrame, dim_name: str, verbose=0):
    values = sorted(dim_df[dim_name].unique())

    # create an "eye" dataframe
    ext_df = pd.DataFrame(data=np.eye(len(values)), columns=values)
    ext_df[dim_name] = values

    join_df = pd.merge(dim_df, ext_df, on=[dim_name])
    join_df = join_df.sort_values(list(dim_df.columns))
    vals = csc_matrix(join_df[values].values)
    if verbose > 0:
        print(values, vals.shape)
    return vals, values


def segment_defs(dim_df: pd.DataFrame, used_dims, verbose=0) -> List[Dict[str, str]]:
    col_defs = []
    this_df = dim_df[used_dims].drop_duplicates().reset_index(drop=True)
    # create an "eye" dataframe on the unique reduced dimensions
    for i, vals in enumerate(this_df.itertuples(index=False)):
        col_defs.append(dict(zip(used_dims, vals)))

    if verbose > 0:
        print(used_dims, len(col_defs))
    return col_defs


def construct_dummies(
    segment_defs: List[Dict[str, str]], cache: Dict[str, Dict[str, np.ndarray]]
) -> scipy.sparse.csc_matrix:
    dummies = []
    for sgdf in segment_defs:
        tmp = None
        for k, v in sgdf.items():
            if tmp is None:
                tmp = cache[k][v]
            else:
                tmp = tmp.multiply(cache[k][v])
        dummies.append(tmp)
    return hstack(dummies)


# This approach was way slower than the join one; keeping it here for reference :)
# def join_to_sparse(dim_df, this_df, chunk_size=100, verbose=0):
#     mats = []
#     tuples = []
#     col_defs = []
#
#     these_dims = list(this_df.columns)
#     # create an "eye" dataframe on the unique reduced dimensions
#     for i, vals in enumerate(this_df.itertuples(index=False)):
#         tuples.append(vals)
#         col_defs.append({k: v for k, v in zip(these_dims, vals)})
#
#     # join it against the real thing, one chunk at a time
#     for i in range(0, len(tuples), chunk_size):
#         this_df = dim_df.copy()
#         these_cols = []
#         for tpl in tuples[i : min(i + chunk_size, len(tuples))]:
#             col_name = "_".join(map(str, tpl))
#             these_cols.append(col_name)
#             this_df[col_name] = 0.0
#
#             for i, (col, value) in enumerate(zip(these_dims, tpl)):
#                 if i == 0:
#                     filter = this_df[col] == value
#                 else:
#                     filter = filter & (this_df[col] == value)
#
#             this_df[filter] = 1.0
#
#         vals = csc_matrix(this_df[these_cols].values)
#         del this_df
#
#         mats.append(vals)
#         # print(vals.shape)
#     if len(mats) > 1:
#         out = hstack(mats)
#     else:
#         out = mats[0]
#     if verbose > 0:
#         print(these_dims, out.shape)
#     return out, col_defs


def sparse_dummy_matrix(
    dim_df: pd.DataFrame,
    min_depth: int = 1,
    max_depth: int = 2,
    verbose=0,
    force_dim: Optional[str] = None,
    clusters: Optional[Dict[str, Sequence[str]]] = None,
    cluster_names: Optional[Dict[str, str]] = None,
    time_basis: Optional[pd.DataFrame] = None,
    max_out_size: int = 1e8,  # threshold num of elements in out matrix
):
    # generate a sparse dummy matrix based on all the combinations
    if force_dim is None:
        dims = list(dim_df.columns)
    else:
        assert force_dim in dim_df.columns
        dims = [c for c in dim_df.columns if c != force_dim]

    if clusters is None:
        clusters = defaultdict(list)

    # drop dimensions with only one value, for clarity
    dims = [d for d in dims if len(dim_df[d].unique()) > 1]

    dims_range_min = min(len(dims), max(1, min_depth))
    dims_range_max = min(len(dims) + 1, max_depth + 1)
    dims_range = range(dims_range_min, dims_range_max)

    # first pass: generate single-dim dummies
    dummy_cache = {}
    for d in dim_df.columns:
        this_mat, these_defs = join_to_sparse(dim_df, d, verbose=verbose)
        dummy_cache[d] = {this_def: this_mat[:, i : i + 1] for i, this_def in enumerate(these_defs)}

    dims_dict = {dim: list(dim_df[dim].unique()) + list(clusters[dim]) for dim in dim_df.columns}

    defs = []
    mats = []

    # Add raw time vectors
    if time_basis is not None:
        for b_name, b_mat in time_basis.items():
            defs.append({"time": b_name})
            mats.append(b_mat)

    # Go over all possible depths
    for num_dims in tqdm(dims_range) if verbose else dims_range:
        # for each depth, sample the possible dimension combinations
        for these_dims in itertools.combinations(dims, num_dims):
            if verbose:
                print(f"Processing {these_dims}")
            if num_dims == 1 and these_dims[0] == "Change from":
                continue
            if force_dim is None:
                used_dims = list(these_dims)
            else:
                used_dims = [force_dim] + list(these_dims)

            segment_constraints = segment_defs_new(dims_dict, used_dims)
            this_mat, these_defs = construct_dummies_new(used_dims, segment_constraints, dummy_cache, cluster_names)
            assert len(these_defs) == this_mat.shape[1]

            if time_basis is None:
                mats.append(this_mat)
                defs += these_defs
            else:
                for b_name, b_mat in time_basis.items():
                    # Multiply the dummies by the time profile
                    if verbose:
                        print(f"Processing {b_name}")
                    re_defs = copy.deepcopy(these_defs)
                    for d in re_defs:
                        d["time"] = b_name

                    # let's split it even deeper to deal with very wide matrices
                    step = math.ceil(max_out_size / b_mat.shape[0])
                    for i in range(0, len(re_defs), step):
                        end_ind = min(i + step, len(re_defs))
                        defs_slice = re_defs[i:end_ind]
                        mat_slice = this_mat[:, i:end_ind]
                        re_mat = diags(b_mat.A.flatten()) @ mat_slice
                        re_mat = csc_matrix(re_mat)
                        assert len(defs_slice) == re_mat.shape[1]

                        mats.append(re_mat)
                        defs += defs_slice

                        test_size = len(defs) * mats[0].shape[0]
                        if test_size >= max_out_size:
                            if verbose:
                                print(f"Threshold reached at {test_size}, dumping")
                            mat = hstack(mats)
                            assert len(defs) == mat.shape[1]
                            yield mat, defs
                            defs = []
                            mats = []
            if len(mats):
                test_size = len(defs) * mats[0].shape[0]
                if test_size >= max_out_size:
                    if verbose:
                        print(f"Threshold reached at {test_size}, dumping")
                    mat = hstack(mats)
                    assert len(defs) == mat.shape[1]
                    yield mat, defs
                    defs = []
                    mats = []
    # mop up
    if len(defs):
        mat = hstack(mats)
        assert len(defs) == mat.shape[1]
        yield mat, defs
    else:
        yield None, None


def segment_defs_new(dims_dict: Dict[str, Sequence[str]], used_dims: List[str]) -> np.ndarray:
    # Look at all possible combinations of dimension values for the chosen dimensions
    if len(used_dims) == 1:
        return np.array(dims_dict[used_dims[0]]).reshape(-1, 1)
    else:
        tmp = segment_defs_new(dims_dict, used_dims[:-1])
        this_dim_values = np.array(dims_dict[used_dims[-1]])
        repeated_values = np.tile(this_dim_values.reshape(-1, 1), len(tmp)).reshape(-1, 1)
        pre_out = np.tile(tmp, (len(this_dim_values), 1))
        out = np.concatenate([pre_out, repeated_values], axis=1)
        return out


def construct_dummies_new(
    used_dims: List[str],
    segment_defs: np.ndarray,
    cache: Dict[str, Dict[str, np.ndarray]],
    cluster_names: Optional[Dict[str, str]] = None,
) -> scipy.sparse.csc_matrix:
    dummies = []
    segments = []
    for sgdf in segment_defs:
        tmp = None
        for i, d in enumerate(used_dims):
            if isinstance(sgdf[i], str) and sgdf[i] not in cache[d]:  # a group of multiple values from that dim
                sub_values = cluster_names[sgdf[i]].split("@@")
                this_dummy = 0
                for val in sub_values:
                    this_dummy += cache[d][val]

            else:
                this_dummy = cache[d][sgdf[i]]

            if tmp is None:
                tmp = this_dummy
            else:
                tmp = tmp.multiply(this_dummy)
        if tmp.sum() > 0:
            dummies.append(tmp)
            segments.append(dict(zip(used_dims, sgdf)))
    return hstack(dummies), segments



from typing import List
from abc import ABC, abstractmethod

import numpy as np


class Fitter(ABC):
    @abstractmethod
    def fit(self, X, y, sample_weight=None):
        pass

    @abstractmethod
    def predict(self, X):
        pass

    def fit_predict(self, X, y, sample_weight=None):
        self.fit(X, y, sample_weight)
        return self.predict(X)

    def error(self, X, y, sample_weight=None):
        # Error is chosen so that it's minimized by the weighted mean of y
        err = y - self.predict(X)
        errsq = err**2
        if sample_weight is not None:
            errsq *= sample_weight
        return np.nansum(errsq)


class AverageFitter(Fitter):
    def __init__(self):
        self.avg = None

    def fit(self, X, y, sample_weight=None):
        y = np.array(y)
        sample_weight = np.array(sample_weight)
        if sample_weight is None:
            self.avg = np.nanmean(y)
        else:
            self.avg = np.nansum(y * sample_weight) / np.nansum(sample_weight)

    def predict(self, X):
        return np.full(X.shape[0], self.avg)



from typing import List, Dict, Tuple
from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.preprocessing import PowerTransformer
from sklearn.cluster import KMeans, kmeans_plusplus
from sklearn.metrics import silhouette_score


def guided_kmeans(X: np.ndarray, power_transform: bool = True) -> np.ndarray:
    """
    Cluster segment averages to calculate aggregated segments
    @param X: Segment mean minus global mean, for each dimension value
    @param power_transform: Do we power transform before clustering
    @return: cluster labels and the transformed values
    """
    if isinstance(X, pd.Series):
        X = X.values.reshape(-1, 1)
    elif isinstance(X, pd.DataFrame):
        X = X.values

    if power_transform:
        if len(X[X > 0]) > 1:
            X[X > 0] = (
                PowerTransformer(standardize=False)
                .fit_transform(X[X > 0].reshape(-1, 1))
                .reshape(-1)
            )
        if len(X[X < 0]) > 1:
            X[X < 0] = (
                -PowerTransformer(standardize=False)
                .fit_transform(-X[X < 0].reshape(-1, 1))
                .reshape(-1)
            )

    best_score = -1
    best_labels = None
    best_n = -1
    # If we allow 2 clusters, it almost always just splits positive vs negative - boring!
    for n_clusters in range(3, int(len(X) / 2) + 1):
        cluster_labels = KMeans(
            n_clusters=n_clusters, init="k-means++", n_init=10
        ).fit_predict(X)
        score = silhouette_score(X, cluster_labels)
        # print(n_clusters, score)
        if score > best_score:
            best_score = score
            best_labels = cluster_labels
            best_n = n_clusters

    # print(best_n)
    return best_labels, X


def to_matrix(labels: np.ndarray) -> np.ndarray:
    out = np.zeros((len(labels), len(labels.unique())))
    for i in labels.unique():
        out[labels == i, i] = 1.0
    return out


def make_clusters(dim_df: pd.DataFrame, dims: List[str]):
    cluster_names = {}
    for dim in dims:
        if len(dim_df[dim].unique()) >= 6:  # otherwise what's the point in clustering?
            grouped_df = (
                dim_df[[dim, "totals", "weights"]].groupby(dim, as_index=False).sum()
            )
            grouped_df["avg"] = grouped_df["totals"] / grouped_df["weights"]
            grouped_df["cluster"], _ = guided_kmeans(grouped_df["avg"])
            pre_clusters = (
                grouped_df[["cluster", dim]]
                .groupby("cluster")
                .agg({dim: lambda x: "@@".join(x)})
                .values
            )
            # filter out clusters with only one element
            these_clusters = [c for c in pre_clusters.reshape(-1) if "@@" in c]
            # create short cluster names
            for i, c in enumerate(these_clusters):
                cluster_names[f"{dim}_cluster_{i + 1}"] = c
    return cluster_names


def nice_cluster_names(x: List[Dict[str, List[str]]]) -> Tuple[List[Dict], Dict]:
    # first pass just populate cluster names
    cluster_strings = defaultdict(set)
    for xx in x:
        for dim, v in xx.items():
            if len(v) > 1:
                cluster_strings[dim].add("@@".join(v))

    cluster_names = {}
    reverse_cluster_names = {}
    for dim, clusters in cluster_strings.items():
        reverse_cluster_names[dim] = {}
        for i, c in enumerate(clusters):
            cluster_names[f"{dim}_cluster_{i + 1}"] = c
            reverse_cluster_names[dim][c] = f"{dim}_cluster_{i + 1}"

    col_defs = []
    for xx in x:
        this_def = {}
        for dim, v in xx.items():
            if len(v) > 1:
                this_def[dim] = reverse_cluster_names[dim]["@@".join(v)]
            else:
                this_def[dim] = v[0]
        col_defs.append(this_def)

    return col_defs, cluster_names



import copy
import json
import warnings
from typing import Optional, Union, List, Dict, Sequence
from collections import defaultdict

import numpy as np
import pandas as pd
from scipy.sparse import csc_matrix, diags

from wise_pizza.solve.find_alpha import find_alpha
from wise_pizza.utils import clean_up_min_max
from wise_pizza.make_matrix import sparse_dummy_matrix
from wise_pizza.cluster import make_clusters
from wise_pizza.preselect import HeuristicSelector
from wise_pizza.time import extend_dataframe
from wise_pizza.slicer_facades import SliceFinderPredictFacade
from wise_pizza.solve.tree import tree_solver
from wise_pizza.solve.solver import solve_lasso


def _summary(obj) -> str:
    out = {"task": obj.task, "segments": obj.segments}
    return json.dumps(out)


class SliceFinder:
    """
    SliceFinder class to find unusual slices
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.min_depth = kwargs.get("min_depth", None)
        self.max_depth = kwargs.get("max_depth", None)
        self.dims = None
        self.X = None
        self.col_defs = None
        self.reg = None
        self.nonzeros = None
        self.weights = None
        self.verbose = 0
        self.task = ""

    def _init_mat(
        self,
        dim_df: pd.DataFrame,
        min_depth: int,
        max_depth: int,
        max_cols: int = 300,
        force_dim: Optional[str] = None,
        clusters: Optional[Dict[str, Sequence[str]]] = None,
        time_basis: Optional[pd.DataFrame] = None,
    ):
        """
        Function to initialize sparse matrix
        @param dim_df: Dataset with dimensions
        @param min_depth: Minimum number of dimension to constrain in segment definition
        @param max_depth: Maximum number of dimension to constrain in segment definition
        @param max_cols: Maxumum number of segments to consider
        @param force_dim: To add dim
        @param clusters: groups of same-dimension values to be considered as candidate segments
        @param time_basis: the set of time profiles to scale the candidate segments by
        @return:
        """
        sel = HeuristicSelector(
            max_cols=max_cols,
            weights=self.weights,
            totals=self.totals,
            time_basis=time_basis,
            verbose=self.verbose,
        )

        # This returns the candidate vectors in batches
        basis_iter = sparse_dummy_matrix(
            dim_df,
            min_depth=min_depth,
            max_depth=max_depth,
            verbose=self.verbose,
            force_dim=force_dim,
            clusters=clusters,
            cluster_names=self.cluster_names,
            time_basis=time_basis,
        )

        # do pre-filter recursively
        for this_X, these_col_defs in basis_iter:
            if this_X is not None:
                X_out, col_defs_out = sel(this_X, these_col_defs)

        if self.verbose:
            print("Preselection done!")
        return X_out, col_defs_out

    def fit(
        self,
        dim_df: pd.DataFrame,
        totals: pd.Series,
        weights: pd.Series = None,
        time_col: pd.Series = None,
        time_basis: pd.DataFrame = None,
        min_segments: int = None,
        max_segments: int = None,
        min_depth: int = 1,
        max_depth: int = 3,
        solver: str = "lp",
        verbose: Union[bool, None] = None,
        force_dim: Optional[str] = None,
        force_add_up: bool = False,
        constrain_signs: bool = True,
        cluster_values: bool = True,
    ):
        """
        Function to fit slicer and find segments
        @param dim_df: Dataset with dimensions
        @param totals: Column with totals
        @param weights: Column with sizes
        @param min_segments: Minimum number of segments to find
        @param max_segments: Maximum number of segments to find, defaults to min_segments
        @param min_depth: Minimum number of dimension to constrain in segment definition
        @param max_depth: Maximum number of dimensions to constrain in segment definition; also max depth pf tree in tree solver
        @param solver: Valid values are "lasso" (default), "tree" (for non-overlapping segments), "omp", or "lp"
        @param verbose: If set to a truish value, lots of debug info is printed to console
        @param force_dim: To add dim
        @param force_add_up: To force add up
        @param constrain_signs: To constrain signs
        @param cluster_values In addition to single-value slices, consider slices that consist of a
        group of segments from the same dimension with similar naive averages

        """

        assert solver.lower() in ["lasso", "tree", "omp", "lp"]
        min_segments, max_segments = clean_up_min_max(min_segments, max_segments)
        if verbose is not None:
            self.verbose = verbose

        totals = np.array(totals).astype(np.float64)

        if weights is None:
            weights = np.ones_like(totals)
        else:
            weights = np.array(weights).astype(np.float64)

        assert min(weights) >= 0
        assert np.sum(np.abs(totals[weights == 0])) == 0

        # Cast all dimension values to strings
        dim_df = dim_df.astype(str)

        dims = list(dim_df.columns)
        # sort the dataframe by dimension values,
        # making sure the other vectors stay aligned
        dim_df = dim_df.reset_index(drop=True)
        dim_df["totals"] = totals
        dim_df["weights"] = weights

        if time_col is not None:
            dim_df["__time"] = time_col
            dim_df = pd.merge(dim_df, time_basis, left_on="__time", right_index=True)
            sort_dims = dims + ["__time"]
        else:
            sort_dims = dims

        dim_df = dim_df.sort_values(sort_dims)
        dim_df = dim_df[dim_df["weights"] > 0]

        # Transform the time basis from table by date to matrices by dataset row
        if time_col is not None:
            self.basis_df = time_basis
            self.time_basis = {}
            for c in time_basis.columns:
                this_ts = dim_df[c].values.reshape((-1, 1))
                max_val = np.abs(this_ts).max()
                # take all the values a nudge away from zero so we can divide by them later
                this_ts[np.abs(this_ts) < 1e-6 * max_val] = 1e-6 * max_val
                self.time_basis[c] = csc_matrix(this_ts)
            self.time = dim_df["__time"].values
        else:
            self.time_basis = None

        self.weights = dim_df["weights"].values
        self.totals = dim_df["totals"].values

        # While we still have weights and totals as part of the dataframe, let's produce clusters
        # of dimension values with similar outcomes
        clusters = defaultdict(list)
        self.cluster_names = {}

        if solver == "tree":
            if cluster_values:
                warnings.warn(
                    "Ignoring cluster_values argument as tree solver makes its own clusters"
                )
            self.X, self.col_defs, self.cluster_names = tree_solver(
                dim_df=dim_df,
                dims=dims,
                time_basis=self.time_basis,
                num_leaves=max_segments,
                max_depth=max_depth,
            )
            self.nonzeros = np.array(range(self.X.shape[1]))
            Xw = csc_matrix(diags(self.weights) @ self.X)
            self.reg = solve_lasso(
                Xw.toarray(),
                self.totals,
                alpha=1e-5,
                verbose=self.verbose,
                fit_intercept=False,
            )
            print("")
        else:
            if cluster_values:
                self.cluster_names = make_clusters(dim_df, dims)
                for dim in dims:
                    clusters[dim] = [
                        c for c in self.cluster_names.keys() if c.startswith(dim)
                    ]

            dim_df = dim_df[dims]  # if time_col is None else dims + ["__time"]]
            self.dim_df = dim_df
            # lazy calculation of the dummy matrix (calculation can be very slow)
            if (
                list(dim_df.columns) != self.dims
                or max_depth != self.max_depth
                or self.X is not None
                and len(dim_df) != self.X.shape[1]
            ):
                self.X, self.col_defs = self._init_mat(
                    dim_df,
                    min_depth,
                    max_depth,
                    force_dim=force_dim,
                    clusters=clusters,
                    time_basis=self.time_basis,
                )
                assert len(self.col_defs) == self.X.shape[1]
                self.min_depth = min_depth
                self.max_depth = max_depth
                self.dims = list(dim_df.columns)

            Xw = csc_matrix(diags(self.weights) @ self.X)

            if self.verbose:
                print("Starting solve!")
            self.reg, self.nonzeros = find_alpha(
                Xw,
                self.totals,
                max_nonzeros=max_segments,
                solver=solver,
                min_nonzeros=min_segments,
                verbose=self.verbose,
                adding_up_regularizer=force_add_up,
                constrain_signs=constrain_signs,
            )

        if self.verbose:
            print("Solver done!!")

        self.segments = [
            {"segment": self.col_defs[i], "index": int(i)} for i in self.nonzeros
        ]

        wgts = np.array((np.abs(Xw[:, self.nonzeros]) > 0).sum(axis=0))[0]

        for i, s in enumerate(self.segments):
            segment_def = s["segment"]
            this_vec = (
                self.X[:, s["index"]]
                .toarray()
                .reshape(
                    -1,
                )
            )
            if "time" in segment_def:
                # Divide out the time profile mult - we've made sure it's always nonzero
                time_mult = (
                    self.time_basis[segment_def["time"]]
                    .toarray()
                    .reshape(
                        -1,
                    )
                )
                dummy = (this_vec / time_mult).astype(int).astype(np.float64)
            else:
                dummy = this_vec

            this_wgts = self.weights * dummy
            wgt = this_wgts.sum()
            # assert wgt == wgts[i]
            s["orig_i"] = i
            s["coef"] = self.reg.coef_[i]
            # TODO: does not taking the abs of coef here break time series?
            s["impact"] = s["coef"] * (np.abs(this_vec) * self.weights).sum()
            s["avg_impact"] = s["impact"] / sum(self.weights)
            s["total"] = (self.totals * dummy).sum()
            s["seg_size"] = wgt
            s["naive_avg"] = s["total"] / wgt

        if time_basis is not None:  # it's a time series product
            # Do we need this bit at all?
            predict = self.reg.predict(self.X[:, self.nonzeros]).reshape(
                -1,
            )
            davg = (predict * self.weights).sum() / self.weights.sum()
            self.reg.intercept_ = -davg

            # And this is the version to use later in TS plotting
            self.predict_totals = self.reg.predict(Xw[:, self.nonzeros]).reshape(
                -1,
            )

            # self.enrich_segments_with_timeless_reg(self.segments, self.y_adj)

        self.segments = self.order_segments(self.segments)

        # In some cases (mostly in a/b exps we have a situation where there is no any diff in totals/sizes)
        if len(self.segments) == 0:
            self.segments.append(
                {
                    "segment": {"No unusual segments": "No unusual segments"},
                    "coef": 0,
                    "impact": 0,
                    "avg_impact": 0,
                    "total": 0,
                    "seg_size": 0,
                    "naive_avg": 0,
                }
            )

    # def enrich_segments_with_timeless_reg(self, segments, y):
    #     pre_X = [s["dummy"] for s in segments]
    #     X = np.concatenate()
    #     pass

    @staticmethod
    def order_segments(segments: List[Dict[str, any]]):
        pos_seg = [s for s in segments if s["impact"] > 0]
        neg_seg = [s for s in segments if s["impact"] < 0]

        return sorted(pos_seg, key=lambda x: abs(x["impact"]), reverse=True) + sorted(
            neg_seg, key=lambda x: abs(x["impact"]), reverse=True
        )

    @staticmethod
    def segment_to_str(segment: Dict[str, any]):
        s = {
            k: v
            for k, v in segment.items()
            if k not in ["coef", "impact", "avg_impact"]
        }
        return str(s)

    @property
    def segment_labels(self):
        return [self.segment_to_str(s["segment"]) for s in self.segments]

    def summary(self):
        return _summary(self)

    @property
    def relevant_cluster_names(self):
        relevant_clusters = {}
        for s in self.segments:
            for c in s["segment"].values():
                if c in self.cluster_names:
                    relevant_clusters[c] = self.cluster_names[c].replace("@@", ", ")
        return relevant_clusters

    def segment_impact_on_totals(self, s: Dict) -> np.ndarray:
        return s["seg_total_vec"]

    @property
    def actual_totals(self):
        return self.totals + self.y_adj

    @property
    def predicted_totals(self):
        return self.predict_totals + self.y_adj

    def predict(
        self,
        steps: Optional[int] = None,
        basis: Optional[pd.DataFrame] = None,
        weight_df: Optional[pd.DataFrame] = None,
    ):
        """
        Predict the totals using the given basis
        :param basis: Time profiles going into the future, time as index
        :param weight_df: dataframe with all dimensions and time, plus weight column
        :return:
        """
        if basis is None:
            if weight_df is None:
                if steps is None:
                    steps = 6
            else:
                steps = len(weight_df[self.time_name].unique())
                warnings.warn(
                    "Ignoring steps argument, using weight_df to determine forecast horizon"
                )
            basis = extend_dataframe(self.basis_df, steps)
        else:
            if steps is not None:
                raise ValueError("Can't specify both basis and steps")

        last_ts = self.time.max()
        new_basis = basis[basis.index > last_ts]

        dims = [c for c in self.dim_df.columns if c != "__time"]
        if weight_df is None:
            pre_dim_df = self.dim_df[dims].drop_duplicates()
            pre_dim_df[self.size_name] = 1

            # Do a Cartesian join of the time basis and the dimensions
            b = new_basis.reset_index().rename(columns={"index": self.time_name})
            b["key"] = 1
            pre_dim_df["key"] = 1

            # Perform the merge operation
            new_dim_df = pd.merge(b, pre_dim_df, on="key").drop("key", axis=1)
        else:
            # This branch is as yet untested
            assert self.time_name in weight_df.columns
            for d in dims:
                assert d in weight_df.columns

            new_dim_df = pd.merge(
                new_basis, weight_df, left_index=True, right_on=self.time_name
            )

        # Join the (timeless) averages to these future rows
        new_dim_df = pd.merge(new_dim_df, self.avg_df, on=dims, how="left").rename(
            columns={"avg": "avg_future"}
        )
        # TODO: replace with a simple regression for more plausible baselines
        global_avg = (
            new_dim_df[self.total_name].sum() / new_dim_df[self.size_name].sum()
        )
        new_dim_df["avg_future"] = new_dim_df["avg_future"].fillna(global_avg)

        # Construct the dummies for predicting

        segments = copy.deepcopy(self.segments)
        new_X = np.zeros((len(new_dim_df), len(segments)))

        new_totals = np.zeros(len(new_dim_df))
        for s in segments:
            dummy, Xi = make_dummy(s["segment"], new_dim_df)
            new_X[:, s["orig_i"]] = Xi
            s["dummy"] = np.concatenate([s["dummy"], dummy], axis=0)
            future_impact = new_dim_df[self.size_name].values * Xi * s["coef"]
            new_totals += future_impact
            s["seg_total_vec"] = np.concatenate([s["seg_total_vec"], future_impact])

        # Evaluate the regression
        new_avg = self.reg.predict(new_X)

        # Add in the constant averages and multiply by the weights
        new_totals = (new_avg + new_dim_df["avg_future"].values) * new_dim_df[
            self.size_name
        ].values

        # Return the dataframe with totals and weights
        new_dim_df[self.total_name] = pd.Series(data=new_totals, index=new_dim_df.index)

        out = SliceFinderPredictFacade(self, new_dim_df, segments)
        return out


def make_dummy(segment_def: Dict[str, str], dim_df: pd.DataFrame) -> np.ndarray:
    """
    Function to make dummy vector from segment definition
    @param segment_def: Segment definition
    @param dim_df: Dataset with dimensions
    @return: Dummy vector
    """
    dummy = np.ones((len(dim_df)))
    for k, v in segment_def.items():
        if k != "time":
            dummy = dummy * (dim_df[k] == v).values

    if "time" in segment_def:
        Xi = dummy * dim_df[segment_def["time"]].values
    else:
        raise ValueError("Segments for time series prediction must contain time!")

    assert np.abs(dummy).sum() > 0
    return dummy, Xi


class SlicerPair:
    def __init__(self, s1: SliceFinder, s2: SliceFinder):
        self.s1 = s1
        self.s2 = s2
        self.task = ""

    def summary(self):
        return _summary(self)



import copy
from dataclasses import dataclass
from typing import List, Dict, Any, Optional, Tuple

import numpy as np
import pandas as pd
from IPython.display import Image, display
import plotly.io as pio
from plotly.io import to_image
from plotly import graph_objects as go
from plotly.subplots import make_subplots

from wise_pizza.slicer_plotting import SliceFinderPlottingInterface


@dataclass
class PlotData:
    df: pd.DataFrame
    nonflat_segments: List[Dict[str, Any]]
    global_time_label: str
    total_name: str
    average_name: str
    sub_titles: List[str]


def plot_time(
    sf: SliceFinderPlottingInterface,
    width: int = 1000,
    height: int = 1000,
    average_name: Optional[str] = None,
    plot_is_static: bool = False,
    return_fig: bool = False,
):
    plot_data = preprocess_for_ts_plot(sf, average_name)
    num_rows = len(plot_data.nonflat_segments) + 1
    fig = make_subplots(
        rows=num_rows,
        cols=3,
        subplot_titles=sum(plot_data.sub_titles, []),
        specs=[[{"secondary_y": False}, {"secondary_y": True}, {"secondary_y": True}]]
        * num_rows,
    )

    plot_single_ts(plot_data, fig, col_nums=(3, 2))
    plot_weights(plot_data, fig, col_num=1)

    for i in range(len(fig.layout.annotations)):
        fig.layout.annotations[i].font.size = 10

    fig.update_layout(
        title_text=f"Actuals vs explanation by segment",
        showlegend=True,
        width=width,
        height=height,
    )
    if plot_is_static:
        image_bytes = to_image(fig, format="png", scale=2)
        return Image(image_bytes, height=height, width=width)
    else:
        if return_fig:
            return fig
        else:
            fig.show()


def plot_ts_pair(
    sf1: SliceFinderPlottingInterface,
    sf2: SliceFinderPlottingInterface,
    width,
    height,
    average_name: str = None,
    plot_is_static: bool = False,
    return_fig: bool = False,
    use_fitted_weights: bool = False,
):
    # if use_fitted_weights:
    #     sf = copy.deepcopy(sf)
    #     sf.s2.totals = (sf.s2.totals/sf.s2.weights)*sf.s1.totals
    #     sf.s2.weights = sf.s1.totals

    wgt_plot_data = preprocess_for_ts_plot(sf1, average_name)  # average name correct?
    totals_plot_data = preprocess_for_ts_plot(sf2, average_name)
    num_rows = max(
        len(wgt_plot_data.nonflat_segments) + 1,
        len(totals_plot_data.nonflat_segments) + 1,
    )
    subplot_titles = []
    for i in range(num_rows):
        if i < len(wgt_plot_data.sub_titles):
            # Totals from weights regression
            subplot_titles.append(wgt_plot_data.sub_titles[i][2])
        else:
            subplot_titles.append("")
        if i < len(totals_plot_data.sub_titles):
            subplot_titles.append(totals_plot_data.sub_titles[i][1])
            subplot_titles.append(totals_plot_data.sub_titles[i][2])
        else:
            subplot_titles.append("")
            subplot_titles.append("")

    fig = make_subplots(
        rows=num_rows,
        cols=3,
        subplot_titles=subplot_titles,
        specs=[[{"secondary_y": True}] * 3] * num_rows,
    )
    min_impacts1, max_impacts1 = plot_single_ts(
        wgt_plot_data, fig, col_nums=(1, None), showlegend=False
    )  # 1, None
    min_impacts32, max_impacts32 = plot_single_ts(
        totals_plot_data, fig, col_nums=(3, 2)
    )  # 3,2

    min_impacts1 = zip(*min_impacts1)
    min_impact_final1 = [min(dim) for dim in min_impacts1]
    max_impacts1 = zip(*max_impacts1)
    max_impact_final1 = [max(dim) for dim in max_impacts1]

    min_impacts32 = zip(*min_impacts32)
    min_impact_final32 = [min(dim) for dim in min_impacts32]
    max_impacts32 = zip(*max_impacts32)
    max_impact_final32 = [max(dim) for dim in max_impacts32]

    for i in range(len(fig.layout.annotations)):
        fig.layout.annotations[i].font.size = 10

    fig.update_layout(
        title_text=f"Actuals vs explanation by segment",
        showlegend=True,
        width=width,
        height=height,
    )

    # # Update secondary y-axis for each column
    for row in range(1, num_rows + 1):
        fig.update_yaxes(
            range=[min_impact_final1[0], max_impact_final1[0]],
            secondary_y=True,
            row=row,
            col=1,
        )

    for row in range(1, num_rows + 1):
        fig.update_yaxes(
            range=[min_impact_final32[1], max_impact_final32[1]],
            secondary_y=True,
            row=row,
            col=2,
        )

    for row in range(1, num_rows + 1):
        fig.update_yaxes(
            range=[min_impact_final32[0], max_impact_final32[0]],
            secondary_y=True,
            row=row,
            col=3,
        )
    if plot_is_static:
        image_bytes = to_image(fig, format="png", scale=2)
        return Image(image_bytes, height=height, width=width)
    else:
        if return_fig:
            return fig
        else:
            fig.show()


def plot_single_ts(
    plotdata: PlotData, fig, showlegend: bool = True, col_nums: Tuple[int, int] = (1, 2)
):
    min_impacts = []
    max_impacts = []
    for i, s in enumerate(plotdata.nonflat_segments):
        agg_df = plotdata.df[s["dummy"] == 1.0].groupby("time", as_index=False).sum()
        # Create subplots
        min_impact, max_impact = simple_ts_plot(
            fig,
            agg_df["time"],
            agg_df["totals"],
            agg_df["weights"],
            reg_seg=agg_df[s["plot_segment"]],
            reg_totals=agg_df["Regr totals"],
            row_num=i + 2,
            showlegend=False,
            col_nums=col_nums,
        )
        min_impacts.append(min_impact)
        max_impacts.append(max_impact)

    # Show the actuals for stuff not in segments
    outside = np.abs(sum([s["dummy"] for s in plotdata.nonflat_segments])) < 1e-8

    left = plotdata.df[outside].groupby("time", as_index=False).sum()
    all_data = plotdata.df.groupby("time", as_index=False).sum()
    min_impact, max_impact = simple_ts_plot(
        fig,
        all_data["time"],
        all_data["totals"],
        all_data["weights"],
        reg_seg=all_data["reg_time_profile"],
        reg_totals=all_data["Regr totals"],
        leftover_totals=left["totals"],
        leftover_avgs=left["totals"] / left["weights"],
        row_num=1,
        showlegend=showlegend,
        col_nums=col_nums,
    )
    min_impacts.insert(0, min_impact)
    max_impacts.insert(0, max_impact)

    return min_impacts, max_impacts


def plot_weights(plotdata: PlotData, fig, col_num: int = 1):
    for i, s in enumerate(plotdata.nonflat_segments):
        agg_df = plotdata.df[s["dummy"] == 1.0].groupby("time", as_index=False).sum()
        zeros = np.zeros(len(agg_df))
        # Create subplots
        min_impact, max_impact = simple_ts_plot(
            fig,
            agg_df["time"],
            agg_df["weights"],
            np.ones(len(agg_df)),
            reg_seg=None,
            reg_totals=None,
            row_num=i + 2,
            showlegend=False,
            col_nums=(col_num, None),
        )

    # Show the actuals for stuff not in segments
    outside = np.abs(sum([s["dummy"] for s in plotdata.nonflat_segments])) < 1e-8

    left = plotdata.df[outside].groupby("time", as_index=False).sum()
    all_data = plotdata.df.groupby("time", as_index=False).sum()

    min_impact, max_impact = simple_ts_plot(
        fig,
        all_data["time"],
        all_data["weights"],
        np.ones(len(all_data)),
        reg_seg=None,
        reg_totals=None,
        leftover_totals=left["weights"],
        leftover_avgs=left["totals"] / left["weights"],
        row_num=1,
        showlegend=False,
        col_nums=(col_num, None),
    )


def same_apart_from_time(s1, s2) -> bool:
    return np.sum(np.abs(s1["dummy"] - s2["dummy"])) < 0.5


def preprocess_for_ts_plot(
    sf: SliceFinderPlottingInterface, average_name: Optional[str] = None
) -> PlotData:
    if average_name is None:
        average_name = "Averages"

    df = pd.DataFrame(
        {
            "totals": sf.actual_totals,
            "Regr totals": sf.predicted_totals,
            "weights": sf.weights,
            "time": sf.time,
        }
    )
    df["reg_time_profile"] = 0.0

    # do a pass over the segments, sorting them into time-only and rest
    global_reg = 0.0
    global_time_profile_names = []
    nonflat_segments = []

    # adj_avg = sf.y_adj.sum() / sf.weights.sum()

    # rel_adj = sf.y_adj - adj_avg * sf.weights

    for i, s in enumerate(sf.segments):
        # Get the segment definition
        segment_def = s["segment"]
        seg_impact = sf.segment_impact_on_totals(s)

        if len(segment_def) > 1:
            almost_duplicate = False
            for s2 in nonflat_segments:
                if same_apart_from_time(s, s2):
                    almost_duplicate = True
                    df[s2["plot_segment"]] += seg_impact
                    s2["segment"]["time"] = (
                        s2["segment"]["time"] + "," + s["segment"]["time"]
                    )
            if not almost_duplicate:
                # offset the segment by actual averages' difference from the global average
                df[f"Seg {i + 1}"] = seg_impact  # + s["dummy"]*rel_adj
                s["plot_segment"] = f"Seg {i+1}"
                nonflat_segments.append(copy.deepcopy(s))

        elif len(segment_def) == 1:
            # Accumulate all pure time profiles into one
            # TODO: this de-duping is almost the same as above, merge!
            df["reg_time_profile"] += seg_impact
            global_time_profile_names.append(segment_def["time"])

        # Do the difference between segment average and global average, for display

    # now create the plots
    if len(global_time_profile_names):
        global_time_label = ", time:" + ",".join(global_time_profile_names)
    else:
        global_time_label = ""

    seg_names = ["All" + global_time_label] + [
        drop_time(s["segment"]) for s in nonflat_segments
    ]
    sub_titles = [
        [
            f"{sf.size_name} for <br>" + f"{s}",
            f"{average_name} for <br>" + f"{s}",
            f"{sf.total_name} for <br>" + f"{s}",
        ]
        if s != "All" and s != global_time_label
        else [
            f"{sf.size_name} for " + "<br>".join([key for key in s]),
            f"{average_name} for " + "<br>".join([key for key in s]),
            f"{sf.total_name} for " + "<br>".join([key for key in s]),
        ]
        for s in seg_names
    ]
    # sub_titles = sum(sub_titles, start=[])

    plot_data = PlotData(
        df, nonflat_segments, global_time_label, sf.total_name, average_name, sub_titles
    )
    return plot_data


def drop_time(s: Dict[str, Any]) -> Dict[str, Any]:
    s = copy.deepcopy(s)
    s.pop("time")
    return s


def naive_dummy(dim_df, seg_def):
    dummy = np.ones(len(dim_df))
    for col, val in seg_def.items():
        if col == "time":
            continue
        dummy[dim_df[col].values != val] = 0
    return dummy


def simple_ts_plot(
    fig,
    time,
    totals,
    weights,
    reg_totals=None,
    leftover_totals=None,
    leftover_avgs=None,
    reg_seg=None,
    row_num=1,
    showlegend: bool = False,
    col_nums: Tuple[int, int] = (1, 2),
):
    min_impact = []
    max_impact = []
    for col in col_nums:
        if col == col_nums[0]:
            mult = 1.0
        else:
            mult = 1 / weights

        if col is None:
            continue

        fig.add_trace(
            go.Bar(
                x=time,
                y=totals * mult,
                name=f"Actuals",
                marker=dict(color="#9fe870"),
                showlegend=showlegend and col == col_nums[0],
            ),
            row=row_num,
            col=col,
        )
        if reg_totals is not None:
            fig.add_trace(
                go.Scatter(
                    x=time,
                    y=reg_totals * mult,
                    mode="lines",
                    name=f"Regression",
                    line=dict(color="#485cc7"),
                    showlegend=showlegend and col == col_nums[0],
                ),
                row=row_num,
                col=col,
            )
        if reg_seg is not None:

            impact = reg_seg * mult
            min_impact.append(min(impact))
            max_impact.append(max(impact))
            fig.add_trace(
                go.Scatter(
                    x=time,
                    y=impact,
                    mode="lines",
                    name=f"Segment's reg contribution (Right axis)",
                    line=dict(color="darkorange"),
                    showlegend=showlegend and col == col_nums[0],
                ),
                row=row_num,
                col=col,
                secondary_y=True,
            )
        if leftover_totals is not None:
            fig.add_trace(
                go.Bar(
                    x=time,
                    y=leftover_totals if col == 1 else leftover_avgs,
                    name=f"Leftover actuals",
                    marker=dict(color="#ff685f"),
                    showlegend=showlegend and col == col_nums[0],
                ),
                row=row_num,
                col=col,
            )

    return min_impact, max_impact



from abc import ABC, abstractmethod
from typing import Dict, Optional

import numpy as np
import pandas as pd


class SliceFinderPlottingInterface(ABC):
    @property
    @abstractmethod
    def actual_totals(self):
        pass

    @property
    @abstractmethod
    def predicted_totals(self):
        pass

    @property
    @abstractmethod
    def weights(self):
        pass

    @property
    @abstractmethod
    def time(self):
        pass

    @property
    @abstractmethod
    def segments(self):
        pass

    @abstractmethod
    def segment_impact_on_totals(self, s: Dict) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def total_name(self):
        pass

    @property
    @abstractmethod
    def size_name(self):
        pass

    @property
    @abstractmethod
    def predict(
        self,
        steps: Optional[int] = None,
        basis: Optional[pd.DataFrame] = None,
        weight_df: Optional[pd.DataFrame] = None,
    ):
        pass



import copy
from typing import Optional, Dict

import numpy as np
import pandas as pd

from wise_pizza.slicer_plotting import SliceFinderPlottingInterface
from wise_pizza.transform import TransformWithWeights, IdentityTransform
from wise_pizza.plotting_time import plot_time


class NotFittedError(Exception):
    """Exception class to raise if predict is called before fit."""

    pass


class SliceFinderPredictFacade(SliceFinderPlottingInterface):
    def __init__(self, sf: "SliceFinder", predict_df: pd.DataFrame, segments: Dict):
        self.sf = sf
        self.df = predict_df
        self._segments = segments

    def plot(
        self,
        plot_is_static=False,
        width=1200,
        height=2000,
        return_fig=False,
        average_name=None,
    ):
        plot_time(
            self,
            # plot_is_static=plot_is_static,
            width=width,
            height=height,
            # return_fig=return_fig,
            average_name=average_name,
        )

    @property
    def actual_totals(self):
        return np.concatenate(
            [self.sf.actual_totals, np.nan * np.zeros((len(self.df)))]
        )

    @property
    def predicted_totals(self):
        return np.concatenate(
            [self.sf.predicted_totals, self.df[self.sf.total_name].values]
        )

    @property
    def actual_avg(self):
        return self.actual_totals / self.weights

    @property
    def predicted_avg(self):
        return self.predicted_totals / self.weights

    @property
    def weights(self):
        return np.concatenate([self.sf.weights, self.df[self.sf.size_name].values])

    @property
    def segments(self):
        return self._segments

    # @property
    # def y_adj(self):
    #     return np.concatenate(
    #         [
    #             self.sf.y_adj,
    #             self.df["avg_future"].values * self.df[self.sf.size_name].values,
    #         ]
    #     )

    @property
    def time(self):
        # This is the column with the time period for each dataset row
        return np.concatenate([self.sf.time, self.df[self.sf.time_name].values])

    @property
    def total_name(self):
        return self.sf.total_name

    @property
    def size_name(self):
        return self.sf.size_name

    def segment_impact_on_totals(self, s: Dict) -> np.ndarray:
        return s["seg_total_vec"]

    def predict(
        self,
        steps: Optional[int] = None,
        basis: Optional[pd.DataFrame] = None,
        weight_df: Optional[pd.DataFrame] = None,
    ):
        raise NotImplementedError("Can't predict on a prediction")

    @property
    def predicted_df(self):
        return self.df


class TransformedSliceFinder(SliceFinderPlottingInterface):
    def __init__(
        self,
        sf: SliceFinderPlottingInterface,
        transformer: Optional[TransformWithWeights] = None,
    ):
        # For now, just use log(1+x) as transform, assume sf was fitted on transformed data
        self.sf = sf
        if transformer is None:
            self.tf = IdentityTransform()
        else:
            self.tf = transformer

        trans_avg = sf.actual_totals / sf.weights  # averages in the transformed space
        trans_pred_avg = sf.predicted_totals / sf.weights

        self.actual_avg = self.tf.inverse_transform_mean(trans_avg)  # a_i
        self.predicted_avg = self.tf.inverse_transform_mean(trans_pred_avg)

        # When we do prediction, the corresponding actuals are nans, so need to use predictions
        # to fill in the gaps for inverse-transforming weights
        patched_avg = copy.deepcopy(trans_avg)
        patched_avg[np.isnan(patched_avg)] = trans_pred_avg[np.isnan(patched_avg)]
        self._weights = self.tf.inverse_transform_weight(sf.weights, patched_avg)

        # probably because of some convexity effect of the exp,
        # predictions end up a touch too high on average post-inverse transform
        # So let's introduce a scaling factor to fix that
        # Can't use self.predicted_totals here, because it needs self.pred_scaler
        predicted_totals = self.predicted_avg * self.weights
        # TODO: debug this!
        # assert not np.isnan(np.sum(predicted_totals))

        nice = ~np.isnan(self.actual_totals * predicted_totals)
        self.pred_scaler = np.sum(self.actual_totals[nice]) / np.sum(
            predicted_totals[nice]
        )

        # Now let's make sure single-segment impacts add up to total impact
        self.segment_mult = 1.0

        self.relevant_cluster_names = sf.relevant_cluster_names

        # Try to make individual segment  impacts add up to total regression post-transform
        # Didn't really make much difference
        # sum_marginals = 0
        # base, _ = self.tf.inverse_transform_totals_weights(self.sf.y_adj, self.sf.weights)
        # pt, _ = self.tf.inverse_transform_totals_weights(self.sf.predicted_totals, self.sf.weights)
        # total_diff = self.pred_scaler*(pt-base)
        #
        # for s in sf.segments:
        #     sum_marginals += self.segment_impact_on_totals(s)
        #
        # self.segment_mult = np.median(np.abs(total_diff/sum_marginals))

    @property
    def actual_totals(self):
        return self.actual_avg * self.weights

    @property
    def predicted_totals(self):
        return self.pred_scaler * self.predicted_avg * self.weights

    @property
    def weights(self):
        return self._weights

    @property
    def segments(self):
        return self.sf.segments

    # @property
    # def y_adj(self):
    #     return self.sf.y_adj

    @property
    def time(self):
        return self.sf.time

    @property
    def total_name(self):
        return self.sf.total_name

    @property
    def size_name(self):
        return self.sf.size_name

    def segment_impact_on_totals(self, s: Dict) -> np.ndarray:
        # Calculate transformed prediction with and without the segment
        totals_without_segment = (
            self.sf.predicted_totals - self.sf.segment_impact_on_totals(s)
        )

        # Transform back and subtract
        pt, _ = self.tf.inverse_transform_totals_weights(
            self.sf.predicted_totals, self.sf.weights
        )
        dpt, _ = self.tf.inverse_transform_totals_weights(
            totals_without_segment, self.sf.weights
        )

        return self.pred_scaler * self.segment_mult * (pt - dpt)

    def predict(
        self,
        steps: Optional[int] = None,
        basis: Optional[pd.DataFrame] = None,
        weight_df: Optional[pd.DataFrame] = None,
    ):
        # Just predict in the transformed space, then return same wrapper
        inner_predict = self.sf.predict(steps=steps, basis=basis, weight_df=weight_df)
        out = TransformedSliceFinder(inner_predict, self.tf)
        out.__class__.plot = SliceFinderPredictFacade.plot
        return out

    @property
    def predicted_df(self):
        if isinstance(self.sf, SliceFinderPredictFacade):
            return self.sf.df
        else:
            raise NotFittedError("Call predict method first to create the prediction")

    # TODO: factor this out
    def plot(
        self,
        plot_is_static=False,
        width=1200,
        height=2000,
        return_fig=False,
        average_name=None,
    ):
        plot_time(
            self,
            # plot_is_static=plot_is_static,
            width=width,
            height=height,
            # return_fig=return_fig,
            average_name=average_name,
        )



from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name="wise-pizza",
    version="0.2.6",
    description="A library to find and visualise the most interesting slices in multidimensional data",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Wise",
    url='https://github.com/transferwise/wise-pizza',
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        "ipython",
        "kaleido",
        "numpy",
        "pandas",
        "pytest",
        "plotly",
        "scikit_learn",
        "scipy>=1.8.0",
        "tqdm",
        "cloudpickle",
        "pivottablejs",
        "streamlit==1.32.0"
    ],
    extras_require={
        "test": [
            "flake8",
            "pytest",
            "pytest-cov"
        ],
    },
    packages=find_packages(
        include=[
            'wise_pizza',
            'wise_pizza.*'
        ],
        exclude=['tests*'],
    ),
    entry_points={
        'console_scripts': [
            'run_wise_pizza_streamlit = wise_pizza.run_streamlit_app_entry_point:main',
        ],
    },
    include_package_data=True,
    keywords='wise-pizza',
)



from typing import Union, List, Tuple
import datetime

import numpy as np
import pandas as pd


def create_time_basis(
    time_values: Union[pd.DataFrame, np.ndarray],
    include_breaks: int = 0,
    baseline_dims: int = 1,
):
    if baseline_dims != 1:
        raise NotImplementedError
    if isinstance(time_values, pd.DataFrame):
        time_values = time_values.values

    t = np.sort(np.unique(time_values))
    const = np.ones(len(t))
    linear = np.cumsum(const)
    linear -= linear.mean()  # now orthogonal to const
    col_names = ["Slope"]

    dummies = [linear]

    if include_breaks:
        for i in range(1, len(t)):
            dummy = np.ones(len(t))
            dummy[:i] = 0
            dummies.append(dummy - dummy.mean())
            # TODO: force date format
            col_names.append(f"{t[i]}_step")
            cum_dummy = np.cumsum(dummy)
            dummies.append(cum_dummy - cum_dummy.mean())
            col_names.append(f"{t[i]}_dtrend")

    dummies = np.stack(dummies)
    out = pd.DataFrame(index=t, columns=col_names, data=dummies.T)
    return out


def extend_dataframe(df: pd.DataFrame, N: int, decay: float = 1.0) -> pd.DataFrame:
    df_extended = df.copy()

    # Try to infer the frequency from the original index
    freq = pd.infer_freq(df.index)

    # Check the type of the original index
    index_type = df.index[0].__class__

    for _ in range(N):
        diff = df_extended.iloc[-1] - df_extended.iloc[-2]
        new_row = df_extended.iloc[-1] + decay * diff

        # If the frequency could not be inferred, use the difference of the last two index values
        if freq is None:
            offset = df_extended.index[-1] - df_extended.index[-2]
            new_date = df_extended.index[-1] + offset
        else:
            new_date = df_extended.index[-1] + pd.tseries.frequencies.to_offset(freq)

        # If the original index was of type date, convert the new date to date
        if index_type == datetime.date:
            new_date = new_date.date()

        df_extended.loc[new_date] = new_row

    return df_extended


def add_average_over_time(
    df: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: str,
    time_name: str,
    cartesian: bool = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    avgs = df[dims + [total_name, size_name]].groupby(dims, as_index=False).sum()
    avgs["avg"] = avgs[total_name] / avgs[size_name]
    if cartesian:
        # make sure that the cartesian product of dimension combinations x time is present,
        # without changing the totals
        times = df[[time_name]].groupby(time_name, as_index=False).sum()
        times["key"] = 1
        avgs["key"] = 1
        cartesian_df = pd.merge(avgs, times, on="key").drop(columns=["key"])
        joined = pd.merge(
            df,
            cartesian_df[dims + [time_name]],
            on=dims + [time_name],
            how="right",
        )
        joined[size_name] = joined[size_name].fillna(
            np.nanmean(joined[size_name].values)
        )
        joined[total_name] = joined[total_name].fillna(0.0)
        df = joined

    avgs = df[dims + [total_name, size_name]].groupby(dims, as_index=False).sum()
    avgs["avg"] = avgs[total_name] / avgs[size_name]
    joined = pd.merge(df, avgs[dims + ["avg"]], on=dims)

    joined["total_adjustment"] = joined[size_name] * joined["avg"]
    out = joined[dims + [total_name, size_name, time_name, "total_adjustment"]]
    tmp = out[dims + [total_name, "total_adjustment"]].groupby(dims).sum()
    assert (tmp[total_name] - tmp["total_adjustment"]).abs().sum() < 1e-6 * df[
        total_name
    ].abs().max()
    return out, avgs[dims + ["avg"]]



import numpy as np


def weighted_quantiles(values, quantiles, sample_weight):
    """Compute the weighted quantile of a 1D numpy array."""
    values_ = np.array(values)
    sample_weight_ = np.array(sample_weight)
    nice = ~np.isnan(values) & ~np.isnan(sample_weight)
    if np.any(~nice):
        raise ValueError("Data contains NaNs")
    sorter = np.argsort(values_)
    sorted_values = values_[sorter]
    sorted_weights = sample_weight_[sorter]
    w_quantiles = np.cumsum(sorted_weights) - 0.5 * sorted_weights
    w_quantiles /= np.sum(sorted_weights)

    try:
        return np.interp(quantiles, w_quantiles, sorted_values)
    except Exception as e:
        print(e)
        raise e
